from __future__ import annotations

import builtins
import collections.abc as c
import copy
import dataclasses
import datetime
import enum
import inspect
import json

import unittest.mock
import pickle
import sys

import pytest

import ansible.module_utils.compat.typing as t

from ansible.module_utils._internal._json._profiles import (
    AnsibleProfileJSONDecoder,
    AnsibleProfileJSONEncoder,
    _JSONSerializationProfile,
)

from ansible.module_utils._internal import _messages

from ansible.module_utils._internal._datatag import (
    AnsibleSerializable,
    AnsibleSerializableEnum,
    AnsibleSingletonTagBase,
    AnsibleTaggedObject,
    NotTaggableError,
    _ANSIBLE_ALLOWED_NON_SCALAR_COLLECTION_VAR_TYPES,
    _AnsibleTaggedStr,
    _empty_frozenset,
    _try_get_internal_tags_mapping,
    _EMPTY_INTERNAL_TAGS_MAPPING,
    AnsibleSerializableDataclass,
    AnsibleSerializableWrapper,
    _TAnsibleSerializable,
    _NO_INSTANCE_STORAGE, AnsibleDatatagBase,
    _tag_dataclass_kwargs,
    AnsibleTagHelper,
)
from ansible.module_utils._internal._datatag._tags import Deprecated
from ansible.module_utils.datatag import native_type_name

if sys.version_info >= (3, 9):
    from typing import get_type_hints
else:
    # 3.8 needs the typing_extensions version of get_type_hints for `include_extras`
    from typing_extensions import get_type_hints


@dataclasses.dataclass(**_tag_dataclass_kwargs)
class ExampleSingletonTag(AnsibleSingletonTagBase):
    pass


@dataclasses.dataclass(**_tag_dataclass_kwargs)
class AnotherExampleSingletonTag(AnsibleSingletonTagBase):
    pass


@dataclasses.dataclass(**_tag_dataclass_kwargs)
class ExampleTagWithContent(AnsibleDatatagBase):
    content_str: str


@dataclasses.dataclass(**_tag_dataclass_kwargs)
class ExampleTagThatPreventsPropagation(AnsibleSingletonTagBase):
    def _get_tag_to_propagate(self, src: t.Any, value: object, *, value_type: t.Optional[type] = None) -> t.Self | None:
        return None


class CopyProtocol(t.Protocol):
    def copy(self) -> t.Any:
        """Copy this instance."""


message_instances = [
    _messages.Event(msg="bla", formatted_source_context="sc"),
    _messages.EventChain(msg_reason="a", traceback_reason="b", event=_messages.Event(msg="c")),
    _messages.ErrorSummary(event=_messages.Event(msg="bla", formatted_traceback="tb")),
    _messages.WarningSummary(event=_messages.Event(msg="bla", formatted_source_context="sc", formatted_traceback="tb")),
    _messages.DeprecationSummary(event=_messages.Event(msg="bla", formatted_source_context="sc", formatted_traceback="tb"), version="1.2.3"),
    _messages.PluginInfo(resolved_name='a.b.c', type=_messages.PluginType.MODULE),
    _messages.PluginType.MODULE,
]


def assert_round_trip(original_value, round_tripped_value, via_copy=False):
    assert original_value == round_tripped_value
    assert AnsibleTagHelper.tags(original_value) == AnsibleTagHelper.tags(round_tripped_value)

    if via_copy and type(original_value) is tuple:  # pylint: disable=unidiomatic-typecheck
        # copy.copy/copy.deepcopy significantly complicate the rules for reference equality with tuple, skip the following checks for values sourced that way
        # tuple impl of __copy__ always returns the same instance, __deepcopy__ always returns the same instance if its contents are immutable
        return

    # singleton values should rehydrate as the shared singleton instance, all others should be a new instance
    if isinstance(original_value, (AnsibleSingletonTagBase, enum.Enum)):
        assert original_value is round_tripped_value
    else:
        assert original_value is not round_tripped_value


class AnsibleSerializableSet(AnsibleSerializableWrapper[set]):
    __slots__ = _NO_INSTANCE_STORAGE

    @classmethod
    def _from_dict(cls: t.Type[_TAnsibleSerializable], d: t.Dict[str, t.Any]) -> set:
        return set(d['value'])

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return dict(
            value=list(self._value),
        )


class AnsibleSerializableTuple(AnsibleSerializableWrapper[tuple]):
    __slots__ = _NO_INSTANCE_STORAGE

    @classmethod
    def _from_dict(cls: t.Type[_TAnsibleSerializable], d: t.Dict[str, t.Any]) -> tuple:
        return tuple(d['value'])

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return dict(
            value=list(self._value),
        )


class AnsibleSerializableBytes(AnsibleSerializableWrapper[bytes]):
    __slots__ = _NO_INSTANCE_STORAGE

    @classmethod
    def _from_dict(cls: t.Type[_TAnsibleSerializable], d: t.Dict[str, t.Any]) -> bytes:
        return d['value'].encode()

    def _as_dict(self) -> t.Dict[str, t.Any]:
        return dict(
            value=self._value.decode(errors='surrogateescape'),
        )


class RoundTripEverything(_JSONSerializationProfile):
    serialize_map = {}
    allowed_ansible_serializable_types = frozenset(AnsibleSerializable._known_type_map.values())

    @classmethod
    def post_init(cls, **kwargs):
        cls.serialize_map.update(AnsibleSerializableWrapper._wrapped_types)


class RoundTripEverythingEncoder(AnsibleProfileJSONEncoder):
    _profile = RoundTripEverything


class RoundTripEverythingDecoder(AnsibleProfileJSONDecoder):
    _profile = RoundTripEverything


@pytest.mark.parametrize("untaggable_instance", [None, True, False])
def test_silent_untaggable(untaggable_instance):
    post_tag = ExampleSingletonTag().tag(untaggable_instance)

    assert post_tag is untaggable_instance


def test_try_tag() -> None:
    untaggable_value = object()
    taggable_value = "Hello"

    assert ExampleSingletonTag().try_tag(untaggable_value) is untaggable_value
    assert ExampleSingletonTag.is_tagged_on(ExampleSingletonTag().try_tag(taggable_value))


def no_op() -> None:
    """No-op function."""


@pytest.mark.parametrize("untaggable_instance", [object(), no_op])
def test_fatal_untaggable(untaggable_instance):
    with pytest.raises(NotTaggableError):
        ExampleSingletonTag().tag(untaggable_instance)


def test_ensure_empty_mapping_singleton() -> None:
    assert type(_EMPTY_INTERNAL_TAGS_MAPPING)() is _EMPTY_INTERNAL_TAGS_MAPPING
    assert copy.copy(_EMPTY_INTERNAL_TAGS_MAPPING) is _EMPTY_INTERNAL_TAGS_MAPPING
    assert copy.deepcopy(_EMPTY_INTERNAL_TAGS_MAPPING) is _EMPTY_INTERNAL_TAGS_MAPPING


def test_get_tags_mapping_from_magicmock() -> None:
    assert _try_get_internal_tags_mapping(unittest.mock.MagicMock()) is _EMPTY_INTERNAL_TAGS_MAPPING


def test_unexpected_reduce_type() -> None:
    with pytest.raises(TypeError):
        ExampleSingletonTag().tag("")._reduce("str")  # type: ignore


_str_override_method_args: t.Dict[str, t.Tuple[tuple, t.Dict[str, t.Any]]] = {
    'partition': ((' ',), {}),
    'removeprefix': ((' ',), {}),
    'removesuffix': ((' ',), {}),
}


def test_tag_types() -> None:
    value = ExampleSingletonTag().tag(AnotherExampleSingletonTag().tag("hi"))

    assert AnsibleTagHelper.tag_types(value) == {ExampleSingletonTag, AnotherExampleSingletonTag}
    assert AnsibleTagHelper.tag_types("hi") is _empty_frozenset


def test_deprecated_invalid_date_type() -> None:
    with pytest.raises(TypeError):
        Deprecated(msg="test", date=42)  # type: ignore


def test_tag_with_invalid_tag_type() -> None:
    with pytest.raises(TypeError):
        AnsibleTagHelper.tag("", ["not a tag"])  # type: ignore


def test_tag_value_type_specified_untagged() -> None:
    value = AnsibleTagHelper.tag(iter((1, 2, 3)), tuple(), value_type=list)

    assert isinstance(value, list)
    assert value == [1, 2, 3]


@dataclasses.dataclass(frozen=True)
class ContainerTestCase:
    """A test case for a container type, either using or not using a generator to source the value."""

    type_under_test: type
    use_generator: bool


class Later:
    def __init__(self, ns: dict, parent_type: type | None = None):
        self._ns = ns
        self._parent_type = parent_type

    def __getattr__(self, item):
        if item in self._ns or hasattr(self._parent_type, item) or self._ns.get('__annotations__', {}).get(item):
            return item

        raise AttributeError(item)


def _default_id_func(obj: object) -> str:
    if type(obj).__name__ == 'EncryptedString':
        return 'EncryptedString'

    res = str(obj)

    if "genexpr" in res:
        res = "(with generator)"

    return res


if sys.version_info < (3, 10):
    # deprecated: description='ditch zip(strict) polyfill' python_version='3.10'
    def zip(*args, **_kwargs):
        return builtins.zip(*args)


@dataclasses.dataclass(frozen=True)
class ParamDesc:
    names: t.List[str]
    id_func: t.Callable[[t.Any], str] = _default_id_func

    @classmethod
    def get_paramdesc_from_hint(cls, annot) -> ParamDesc:
        if annot:
            for meta in getattr(annot, '__metadata__', []):
                if isinstance(meta, ParamDesc):
                    return meta

        return ParamDesc(names=[])

    @classmethod
    def get_test_param_values(
        cls,
        obj: t.Any,
        name: str,
        maybe_with_names: t.Iterable[str],
    ) -> tuple[t.Sequence[str], t.Iterable[t.Any], t.Callable[[object], str]]:
        value = getattr(obj, name)

        try:
            if inspect.ismethod(value):
                annot = get_type_hints(value, include_extras=True).get('return')
                value = value()
            else:
                annot = get_type_hints(obj, include_extras=True).get(name)
        except Exception as ex:
            raise Exception(f"failed getting type hints for {obj!r} {name!r}") from ex

        paramdesc = cls.get_paramdesc_from_hint(annot)

        if not paramdesc.names:
            paramdesc = dataclasses.replace(paramdesc, names=["value"])

        col_count = len(paramdesc.names)

        if col_count == 1:
            col_count = 0  # HACK: don't require a wrapper container around single-element rows

        maybe_with_names = set(maybe_with_names)

        # simulate ordered set with no-values dict; the output order is not important but must be consistent per-row; use the input data order for now
        matched_names = {n: None for n in paramdesc.names}

        if not matched_names:
            return [], [], str

        out_values = []

        # DTFIX-FUTURE: apply internal tagging/annotation to point at the source data row on test failure/error?
        for rownum, row in enumerate(value or []):
            if col_count:
                # validate column count and filter the args, returning them in `matched_names` order
                if len(row) != col_count:
                    raise ValueError(f"row {rownum} of {name!r} must contain exactly {col_count} value(s); found {len(row)}")

                out_values.append([argvalue for argname, argvalue in zip(paramdesc.names, row, strict=True) if argname in matched_names])
            else:
                # just return the entire row as "value"
                out_values.append([row])

        return list(matched_names), out_values, paramdesc.id_func


class AutoParamSupport:
    def pytest_generate_tests(self, metafunc: pytest.Metafunc):
        for node, mark in metafunc.definition.iter_markers_with_node("autoparam"):
            attrname = mark.args[0]
            test_arg_names = set(inspect.signature(metafunc.function).parameters)
            argnames, argvalues, id_func = ParamDesc.get_test_param_values(metafunc.cls, attrname, test_arg_names)
            if argnames:
                metafunc.parametrize(argnames, argvalues, ids=id_func)

    def __init_subclass__(cls, **kwargs):
        cls.post_init()

    @classmethod
    def post_init(cls) -> None:
        ...


class TestDatatagTarget(AutoParamSupport):
    later = t.cast(t.Self, Later(locals()))

    tag_instances_with_reprs: t.Annotated[t.List[t.Tuple[AnsibleDatatagBase, str]], ParamDesc(["value", "expected_repr"])] = [
        (Deprecated(msg="hi mom, I am deprecated", date='2023-01-02', version="42.42"),
         "Deprecated(msg='hi mom, I am deprecated', date='2023-01-02', version='42.42')"),
        (Deprecated(msg="minimal"), "Deprecated(msg='minimal')")
    ]

    taggable_container_instances: t.List[c.Collection] = [
        dict(hi="mom"),
        ['hi', 'mom'],
        {'hi mom'},  # kept as a single item set to allow repr() testing without worrying about non-deterministic order of set items
        ("hi", "mom",),
    ]

    taggable_instances: t.List[object] = taggable_container_instances + [
        b'hi mom',
        42.0,
        42,
        "hi mom",
        datetime.datetime(2023, 9, 15, 21, 5, 30, 1900, datetime.timezone.utc),
        datetime.date(2023, 9, 15),
        datetime.time(21, 5, 30, 1900),
    ]

    tagged_object_instances: t.List[AnsibleTaggedObject] = [
        t.cast(AnsibleTaggedObject, ExampleTagWithContent(content_str=__file__).tag(item)) for item in taggable_instances]

    datatag_instances: t.List[AnsibleDatatagBase]
    serializable_instances: t.List[object]
    serializable_instances_with_instance_copy: t.List[CopyProtocol]
    serializable_types: t.List[t.Type[AnsibleSerializable]]

    @classmethod
    def post_init(cls) -> None:
        cls.datatag_instances = [value for value, _repr in cls.tag_instances_with_reprs]
        cls.serializable_instances = [value for value in (
            cls.datatag_instances + cls.tagged_object_instances + message_instances) if cls.is_type_applicable(value)]
        cls.serializable_instances_with_instance_copy = [t.cast(CopyProtocol, item) for item in cls.serializable_instances if hasattr(item, 'copy')]
        # NOTE: this doesn't include the lazy template types, those are tested separately
        cls.serializable_types = [value for value in (
            list(AnsibleSerializable._known_type_map.values()) + [AnsibleSerializable]) if cls.is_type_applicable(value)]

    @classmethod
    def is_type_applicable(cls, type_obj) -> bool:
        return (
            type_obj.__module__ != __name__ and  # exclude test-only tags/objects
            cls.is_controller_only_type(type_obj) == cls.is_controller_only_test()
        )

    @staticmethod
    def is_controller_only_type(type_obj: type) -> bool:
        return 'module_utils' not in type_obj.__module__

    @classmethod
    def is_controller_only_test(cls) -> bool:
        return cls.is_controller_only_type(cls)

    @classmethod
    def container_test_cases(cls) -> t.Annotated[t.List[t.Tuple[t.Any, t.Optional[type], type]], ParamDesc(["value", "value_type", "type_under_test"])]:
        """
        Return container test parameters for the given test case.
        Called during each test run to create the test value on-demand.
        """

        test_cases = []

        for test_case in create_container_test_cases(_ANSIBLE_ALLOWED_NON_SCALAR_COLLECTION_VAR_TYPES):
            instances = cls.taggable_instances + cls.tagged_object_instances

            # pylint: disable=unidiomatic-typecheck
            candidates = [instance for instance in instances if type(instance) is test_case.type_under_test]

            assert len(candidates) == 1, f"container_test_parameters found {len(candidates)}, expected 1"

            value = candidates[0]

            test_cases.append(create_container_test_parameters(test_case, value))

        return test_cases

    @pytest.mark.autoparam(later.tag_instances_with_reprs)
    def test_tag_repr(self, value: t.Any, expected_repr: str):
        assert repr(value) == expected_repr

    @pytest.mark.autoparam(later.container_test_cases)
    def test_tag_copy(self, value: t.Collection, value_type: type | None, type_under_test: type) -> None:
        """Ensure copying tags returns the correct type and tags."""
        tag = ExampleSingletonTag()
        src = tag.tag("tagged")

        result: t.Collection = AnsibleTagHelper.tag_copy(src, value, value_type=value_type)

        assert isinstance(result, type_under_test)
        assert tag in AnsibleTagHelper.tags(result)

    @pytest.mark.autoparam(later.taggable_instances)
    @pytest.mark.allow_delazify  # this test requires a working templar on lazies
    def test_dir(self, value: object) -> None:
        """Ensure the dir() of a tagged instance is identical to the dir() returned by the underlying native Python type, excluding `_` prefixed names."""
        tagged_instance = ExampleSingletonTag().tag(value)

        assert tagged_instance is not value
        assert ([name for name in dir(tagged_instance) if not name.startswith('_')] ==
                [name for name in dir(AnsibleTagHelper.as_native_type(tagged_instance)) if not name.startswith('_')])

    @pytest.mark.autoparam(later.taggable_instances)
    @pytest.mark.allow_delazify  # this test requires a working templar on lazies
    def test_repr(self, value: object) -> None:
        """Ensure the repr() of a tagged instance is identical to the repr() returned by the underlying native Python type."""
        tagged_instance = ExampleSingletonTag().tag(value)

        assert tagged_instance is not value
        assert repr(tagged_instance) == repr(value)

    @pytest.mark.autoparam(later.taggable_instances)
    @pytest.mark.allow_delazify  # this test requires a working templar on lazies
    def test_str(self, value: object) -> None:
        """Ensure the str() of a tagged instance is identical to the str() returned by the underlying native Python type."""
        tagged_instance = ExampleSingletonTag().tag(value)

        assert tagged_instance is not value
        assert str(tagged_instance) == str(value)

    def test_serializable_instances_cover_all_concrete_impls(self):
        tested_types = {type(instance_type) for instance_type in self.serializable_instances}

        excluded_type_names = {
            AnsibleTaggedObject.__name__,  # base class, cannot be abstract
            AnsibleSerializableDataclass.__name__,  # base class, cannot be abstract
            AnsibleSerializable.__name__,  # base class, cannot be abstract
            AnsibleSerializableEnum.__name__,  # base class, cannot be abstract
            # these types are all controller-only, so it's easier to have static type names instead of importing them
            'JinjaConstTemplate',  # serialization not required
            '_EncryptedSource',  # serialization not required
            'CapturedErrorSummary',  # serialization not required
        }

        # don't require instances for types marked abstract or types that are clearly intended to be so (but can't be marked as such)
        required_types = {instance_type for instance_type in self.serializable_types if (
            not inspect.isabstract(instance_type) and
            not instance_type.__name__.endswith('Base') and
            'Lazy' not in instance_type.__name__ and  # lazy types use the same input data
            instance_type.__name__ not in excluded_type_names and
            not issubclass(instance_type, AnsibleSerializableWrapper)
        )}

        missing_types = required_types.difference(tested_types)

        assert not missing_types

    @pytest.mark.autoparam(later.serializable_instances)
    @pytest.mark.allow_delazify  # this test requires a working templar on lazies
    def test_json_roundtrip(self, value: object):
        """
        Verify that the serialization infrastructure (profiles) can round-trip any type we choose to support.
        For taggable types which have no production use-case for round-tripping, the necessary wrapper types have been implemented in this test module.
        """
        payload = json.dumps(value, cls=RoundTripEverythingEncoder)
        round_tripped_value = json.loads(payload, cls=RoundTripEverythingDecoder)

        assert_round_trip(value, round_tripped_value)

    @pytest.mark.autoparam(later.serializable_instances)
    def test_pickle_roundtrip(self, value: object):
        if "Lazy" in type(value).__name__:
            pytest.xfail("pickle prohibited on lazies")

        pickled_value = pickle.dumps(value)
        round_tripped_value = pickle.loads(pickled_value)

        assert_round_trip(value, round_tripped_value)

    @pytest.mark.autoparam(later.serializable_instances)
    def test_deepcopy_roundtrip(self, value: object):
        if "Lazy" in type(value).__name__:
            pytest.xfail("deepcopy not supported on lazies yet")

        round_tripped_value = copy.deepcopy(value)

        # DTFIX5: ensure items in collections are copies

        assert_round_trip(value, round_tripped_value, via_copy=True)

    @pytest.mark.autoparam(later.tagged_object_instances)
    def test_native_copy(self, value: AnsibleTaggedObject) -> None:
        native_copy = value._native_copy()

        assert type(value) is not type(native_copy)
        assert isinstance(value, type(native_copy))

        if not isinstance(native_copy, int):
            assert native_copy is not value._native_copy()

        # DTFIX5: ensure items in collections are not copies

        assert native_copy == value
        assert native_copy == value._native_copy()

    @pytest.mark.autoparam(later.serializable_instances)
    def test_copy_roundtrip(self, value: object):
        if "Lazy" in type(value).__name__:
            pytest.xfail("copy prohibited on lazies")

        round_tripped_value = copy.copy(value)

        # DTFIX5: ensure items in collections are not copies

        assert_round_trip(value, round_tripped_value, via_copy=True)

    @pytest.mark.autoparam(later.serializable_instances_with_instance_copy)
    def test_instance_copy_roundtrip(self, value: CopyProtocol):
        round_tripped_value = value.copy()

        # DTFIX5: ensure items in collections are not copies

        assert_round_trip(value, round_tripped_value)

    test_dataclass_tag_base_field_validation_fail_instances: t.Annotated[
        t.List[t.Tuple[t.Type[AnsibleDatatagBase], t.Dict[str, object]]], ParamDesc(["tag_type", "init_kwargs"])
    ] = [
        (Deprecated, dict(msg=ExampleSingletonTag().tag(''))),
        (Deprecated, dict(date=ExampleSingletonTag().tag(''), msg='')),
        (Deprecated, dict(version=ExampleSingletonTag().tag(''), msg='')),
    ]

    @pytest.mark.autoparam(later.test_dataclass_tag_base_field_validation_fail_instances)
    def test_dataclass_tag_base_field_validation_fail(self, tag_type: t.Callable, init_kwargs: t.Dict[str, t.Any]) -> None:
        field_name = list(init_kwargs.keys())[0]
        actual_type = type(init_kwargs[field_name])

        with pytest.raises(TypeError, match=f"{field_name} must be <class '.*'> instead of {actual_type}"):
            tag_type(**init_kwargs)

    test_dataclass_tag_base_field_validation_pass_instances: t.Annotated[
        t.List[t.Tuple[t.Type[AnsibleDatatagBase], t.Dict[str, object]]], ParamDesc(["tag_type", "init_kwargs"])
    ] = [
        (Deprecated, dict(msg='')),
        (Deprecated, dict(msg='', date='2025-01-01')),
        (Deprecated, dict(msg='', version='')),
    ]

    @pytest.mark.autoparam(later.test_dataclass_tag_base_field_validation_pass_instances)
    def test_dataclass_tag_base_field_validation_pass(self, tag_type: t.Callable, init_kwargs: t.Dict[str, t.Any]) -> None:
        tag_type(**init_kwargs)

    @pytest.mark.autoparam(later.taggable_instances)
    @pytest.mark.allow_delazify  # this test requires a working templar on lazies
    def test_as_untagged_type(self, value: object) -> None:
        """
        Ensure that `as_untagged_type` preserves object reference identity for untagged inputs, and
        that tagged inputs are returned as their original native types.
        """
        tagged_instance = ExampleSingletonTag().tag(value)
        roundtripped_instance = AnsibleTagHelper.as_native_type(tagged_instance)

        if not isinstance(value, AnsibleTaggedObject):  # lazies are always a tagged type, so as_untagged_type will be a copy
            assert AnsibleTagHelper.as_native_type(value) is value
            assert type(roundtripped_instance) is type(value)

        assert roundtripped_instance == value

    @pytest.mark.autoparam(later.taggable_instances)
    def test_untag(self, value: object) -> None:
        """Ensure tagging and then untagging a taggable instance returns new instances as appropriate, with the correct tags and type."""
        tagged_instance = ExampleSingletonTag().tag(AnotherExampleSingletonTag().tag(value))

        tags_unchanged = Deprecated.untag(tagged_instance)  # not tagged with this value, nothing to do

        assert tags_unchanged is tagged_instance

        one_less_tag = AnotherExampleSingletonTag.untag(tagged_instance)

        assert one_less_tag is not tagged_instance
        assert type(one_less_tag) is type(tagged_instance)  # pylint: disable=unidiomatic-typecheck
        assert AnsibleTagHelper.tags(one_less_tag) == frozenset((ExampleSingletonTag(),))

        no_tags = ExampleSingletonTag.untag(one_less_tag)

        assert no_tags is not one_less_tag
        assert type(no_tags) is type(value)
        assert AnsibleTagHelper.tags(no_tags) is _empty_frozenset

        still_no_tags = ExampleSingletonTag.untag(no_tags)

        assert still_no_tags is no_tags

    @pytest.mark.autoparam(later.serializable_types)
    def test_slots(self, value: type) -> None:
        """Assert that __slots__ are properly defined on the given serializable type."""
        if value in (AnsibleSerializable, AnsibleTaggedObject):
            expect_slots = True  # non-dataclass base types have no attributes, but still use slots
        elif issubclass(value, (int, bytes, tuple, enum.Enum)):
            # non-empty slots are not supported by these variable-length data types
            # see: https://docs.python.org/3/reference/datamodel.html
            expect_slots = False
        elif issubclass(value, AnsibleSerializableDataclass) or value == AnsibleSerializableDataclass:
            assert dataclasses.is_dataclass(value)  # everything extending AnsibleSerializableDataclass must be a dataclass
            expect_slots = sys.version_info >= (3, 10)  # 3.10+ dataclasses have attributes (and support slots)
        else:
            expect_slots = True  # normal types have attributes (and slots)

        # check for slots on the type itself, ignoring slots on parents
        has_slots = '__slots__' in value.__dict__
        assert has_slots == expect_slots

        # instances of concrete types using __slots__ should not have __dict__ (which would indicate missing __slots__ definitions in the class hierarchy)
        serializable_instance = {type(instance): instance for instance in self.serializable_instances}.get(value)

        if serializable_instance:
            has_dict = hasattr(serializable_instance, '__dict__')
            assert has_dict != expect_slots

    #

    # WORKING
    @pytest.mark.autoparam(later.container_test_cases)
    def test_tag(self, value: object, value_type: type | None, type_under_test: type) -> None:
        """Ensure tagging a value returns the correct type and tags."""

        tag = ExampleSingletonTag()

        result: t.Collection = AnsibleTagHelper.tag(value, tags=tag, value_type=value_type)  # type: ignore[arg-type]

        assert isinstance(result, type_under_test)
        assert tag in AnsibleTagHelper.tags(result)


def create_container_test_parameters(test_case: ContainerTestCase, value: t.Any) -> t.Tuple[t.Any, t.Optional[type], type]:
    """
    Return container test parameters for the given test case and collection instance.
    The result is tuple of three values:
    - 1) The value or a generator.
    - 2) The type represented by the generator, or None when not using a generator.
    - 3) The type represented by the value.
    """
    if test_case.use_generator:
        # This test creates a generator to source items from the value to facilitate optimized creation of collections when tagging and copying tags.
        # To avoid triggering special behavior during iteration, a native copy is used when the value is a tagged object.

        if isinstance(value, AnsibleTaggedObject):
            native_value = value._native_copy()
        else:
            native_value = value

        if isinstance(value, c.Mapping):
            generator = ((k, v) for k, v in native_value.items())
        else:
            generator = (item for item in native_value)

        return generator, test_case.type_under_test, test_case.type_under_test  # testing via a generator, which requires use of value_type

    return value, None, test_case.type_under_test  # testing the actual type without specifying value_type


def create_container_test_cases(types: t.Iterable[t.Type[t.Collection]]) -> t.List[ContainerTestCase]:
    """
    Return a list of test cases for the given types.
    Each type will result in two test cases, one that uses a generator and one that does not.
    """
    sources: list[ContainerTestCase] = []

    for type_under_test in sorted(types, key=lambda item: item.__name__):
        sources.extend((
            ContainerTestCase(type_under_test, False),  # testing the actual type without specifying value_type
            ContainerTestCase(type_under_test, True),  # testing via a generator, which requires use of value_type
        ))

    return sources


@pytest.mark.parametrize("value, expected_type_name", (
    (1, 'int'),
    (ExampleSingletonTag().tag(1), 'int'),
    (str, 'str'),
    (_AnsibleTaggedStr, 'str'),
))
def test_friendly_name(value: object, expected_type_name: str) -> None:
    assert native_type_name(value) == expected_type_name


def test_deserialize_unknown_type() -> None:
    with pytest.raises(ValueError):
        AnsibleSerializable._deserialize({AnsibleSerializable._TYPE_KEY: 'bogus'})


def test_conflicting_tagged_type_map_entry():
    def create_problem():
        class SecondaryDict(dict, AnsibleTaggedObject):
            pass

        return SecondaryDict()  # pragma: nocover

    with pytest.raises(TypeError, match="Cannot define type 'SecondaryDict' since '_AnsibleTaggedDict' already extends 'dict'."):
        create_problem()


@pytest.mark.parametrize("value,expected_idx", (
    (('a', ExampleSingletonTag().tag('b'), ExampleSingletonTag().tag('c')), 1),
    ((ExampleSingletonTag().tag('a'), ExampleSingletonTag().tag('b'), 'c'), 0),
    ((ExampleTagWithContent(content_str='').tag('a'), 'b'), None),
))
def test_first_tagged_on(value: c.Sequence, expected_idx: int | None):
    expected = value[expected_idx] if expected_idx is not None else None

    assert ExampleSingletonTag.first_tagged_on(*value) is expected


class NonNativeTaggedType(AnsibleTaggedObject):
    """A surrogate test type that allows empty tags."""
    __slots__ = ('_ansible_tags_mapping', '_value')
    _empty_tags_as_native: t.ClassVar[bool] = False

    _value: str

    def __init__(self, value: str):
        self._ansible_tags_mapping = _EMPTY_INTERNAL_TAGS_MAPPING


def test_helper_untag():
    """Validate the behavior of the `AnsibleTagHelper.untag` method."""
    value = AnsibleTagHelper.tag("value", tags=[ExampleSingletonTag(), ExampleTagWithContent(content_str="blah")])
    assert len(AnsibleTagHelper.tag_types(value)) == 2

    less_est = AnsibleTagHelper.untag(value, ExampleSingletonTag)
    assert AnsibleTagHelper.tag_types(less_est) == {ExampleTagWithContent}

    no_tags_explicit = AnsibleTagHelper.untag(value, ExampleSingletonTag, ExampleTagWithContent)
    assert type(no_tags_explicit) is str  # pylint: disable=unidiomatic-typecheck

    no_tags_implicit = AnsibleTagHelper.untag(value)
    assert type(no_tags_implicit) is str  # pylint: disable=unidiomatic-typecheck

    untagged_value = "not a tagged value"
    assert AnsibleTagHelper.untag(untagged_value) is untagged_value

    tagged_empty_tags_ok_value = ExampleSingletonTag().tag(NonNativeTaggedType("blah"))
    untagged_empty_tags_ok_value = AnsibleTagHelper.untag(tagged_empty_tags_ok_value)
    assert type(untagged_empty_tags_ok_value) is NonNativeTaggedType  # pylint: disable=unidiomatic-typecheck
    assert not AnsibleTagHelper.tags(untagged_empty_tags_ok_value)


def test_serializable_dataclass_with_tuple() -> None:
    """Validate that dataclass deserialization converts inbound lists for tuple-typed fields."""
    @dataclasses.dataclass(**_tag_dataclass_kwargs)
    class HasTuple(AnsibleSerializableDataclass):
        data: t.Tuple[str, ...]

    assert HasTuple._from_dict(dict(data=["abc", "def"])) == HasTuple(data=("abc", "def"))


def test_tag_copy_non_propagation() -> None:
    value = ExampleTagThatPreventsPropagation().tag("hello")
    copied_value = AnsibleTagHelper.tag_copy(value, "copy")

    assert type(copied_value) is str  # pylint: disable=unidiomatic-typecheck
    assert copied_value == "copy"



from __future__ import annotations

from ansible.module_utils.datatag import deprecator_from_collection_name
from ansible.module_utils.common.warnings import deprecate

# extra lines below to allow for adding more imports without shifting the line numbers of the code that follows
#
#
#
#
#
#
#
#


def do_stuff() -> None:
    deprecator = deprecator_from_collection_name('ns.col')

    # ansible-deprecated-version - only ansible-core can encounter this
    deprecate(msg='ansible-deprecated-no-version', collection_name='ns.col')
    # ansible-invalid-deprecated-version - only ansible-core can encounter this
    deprecate(msg='collection-deprecated-version', collection_name='ns.col', version='1.0.0')
    deprecate(msg='collection-invalid-deprecated-version', collection_name='ns.col', version='not-a-version')
    # ansible-deprecated-no-collection-name - module_utils cannot encounter this
    deprecate(msg='wrong-collection-deprecated', collection_name='ns.wrong', version='3.0.0')
    deprecate(msg='ansible-expired-deprecated-date', collection_name='ns.col', date='2000-01-01')
    deprecate(msg='ansible-invalid-deprecated-date', collection_name='ns.col', date='not-a-date')
    deprecate(msg='ansible-deprecated-both-version-and-date', collection_name='ns.col', version='3.0.0', date='2099-01-01')
    deprecate(msg='removal-version-must-be-major', collection_name='ns.col', version='3.1.0')
    # ansible-deprecated-date-not-permitted - only ansible-core can encounter this
    # ansible-deprecated-unnecessary-collection-name - module_utils cannot encounter this
    # ansible-deprecated-collection-name-not-permitted - only ansible-core can encounter this
    deprecate(msg='ansible-deprecated-both-collection-name-and-deprecator', collection_name='ns.col', deprecator=deprecator, version='3.0.0')



from __future__ import annotations as _annotations

import typing as _t

from ansible._internal._yaml import _loader


def AnsibleLoader(*args, **kwargs) -> _t.Any:
    """Compatibility factory function; returns an Ansible YAML loader instance."""
    return _loader.AnsibleLoader(*args, **kwargs)



from __future__ import annotations

import collections.abc as c


def raise_exceptions(exceptions: c.Sequence[BaseException]) -> None:
    """
    Raise a chain of exceptions from the given exception list.
    Exceptions will be raised starting from the end of the list.
    """
    if len(exceptions) > 1:
        try:
            raise_exceptions(exceptions[1:])
        except Exception as ex:
            raise exceptions[0] from ex

    raise exceptions[0]



from __future__ import annotations

import contextlib
import enum
import typing as t

from ansible.utils.display import Display
from ansible.constants import config

display = Display()

# FUTURE: add sanity test to detect use of skip_on_ignore without Skippable (and vice versa)


class ErrorAction(enum.Enum):
    """Action to take when an error is encountered."""

    IGNORE = enum.auto()
    WARNING = enum.auto()
    ERROR = enum.auto()

    @classmethod
    def from_config(cls, setting: str, variables: dict[str, t.Any] | None = None) -> t.Self:
        """Return an `ErrorAction` enum from the specified Ansible config setting."""
        return cls[config.get_config_value(setting, variables=variables).upper()]


class _SkipException(BaseException):
    """Internal flow control exception for skipping code blocks within a `Skippable` context manager."""

    def __init__(self) -> None:
        super().__init__('Skipping ignored action due to use of `skip_on_ignore`. It is a bug to encounter this message outside of debugging.')


class _SkippableContextManager:
    """Internal context manager to support flow control for skipping code blocks."""

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type, _exc_val, _exc_tb) -> bool:
        if exc_type is None:
            raise RuntimeError('A `Skippable` context manager was entered, but a `skip_on_ignore` handler was never invoked.')

        return exc_type is _SkipException  # only mask a _SkipException, allow all others to raise


Skippable = _SkippableContextManager()
"""Context manager singleton required to enclose `ErrorHandler.handle` invocations when `skip_on_ignore` is `True`."""


class ErrorHandler:
    """
    Provides a configurable error handler context manager for a specific list of exception types.
    Unhandled errors leaving the context manager can be ignored, treated as warnings, or allowed to raise by setting `ErrorAction`.
    """

    def __init__(self, action: ErrorAction) -> None:
        self.action = action

    @contextlib.contextmanager
    def handle(self, *args: type[BaseException], skip_on_ignore: bool = False) -> t.Iterator[None]:
        """
        Handle the specified exception(s) using the defined error action.
        If `skip_on_ignore` is `True`, the body of the context manager will be skipped for `ErrorAction.IGNORE`.
        Use of `skip_on_ignore` requires enclosure within the `Skippable` context manager.
        """
        if not args:
            raise ValueError('At least one exception type is required.')

        if skip_on_ignore and self.action == ErrorAction.IGNORE:
            raise _SkipException()  # skipping ignored action

        try:
            yield
        except args as ex:
            match self.action:
                case ErrorAction.WARNING:
                    display.error_as_warning(msg=None, exception=ex)
                case ErrorAction.ERROR:
                    raise
                case _:  # ErrorAction.IGNORE
                    pass

        if skip_on_ignore:
            raise _SkipException()  # completed skippable action, ensures the `Skippable` context was used

    @classmethod
    def from_config(cls, setting: str, variables: dict[str, t.Any] | None = None) -> t.Self:
        """Return an `ErrorHandler` instance configured using the specified Ansible config setting."""
        return cls(ErrorAction.from_config(setting, variables=variables))



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import collections

from ansible.modules.apt import expand_pkgspec_from_fnmatches
import pytest

FakePackage = collections.namedtuple("Package", ("name",))
fake_cache = [
    FakePackage("apt"),
    FakePackage("apt-utils"),
    FakePackage("not-selected"),
]


@pytest.mark.parametrize(
    ("test_input", "expected"),
    [
        pytest.param(
            ["apt"],
            ["apt"],
            id="trivial",
        ),
        pytest.param(
            ["apt=1.0*"],
            ["apt=1.0*"],
            id="version-wildcard",
        ),
        pytest.param(
            ["apt*=1.0*"],
            ["apt", "apt-utils"],
            id="pkgname-wildcard-version",
        ),
        pytest.param(
            ["apt*"],
            ["apt", "apt-utils"],
            id="pkgname-expands",
        ),
    ],
)
def test_expand_pkgspec_from_fnmatches(test_input, expected):
    """Test positive cases of ``expand_pkgspec_from_fnmatches``."""
    assert expand_pkgspec_from_fnmatches(None, test_input, fake_cache) == expected



from __future__ import annotations

import pytest


def test_failed_non_interactive_captured_command() -> None:
    """Verify failed non-interactive captured commands raise a `SubprocessError` with `stdout` and `stderr` set."""
    from ansible_test._internal.util import raw_command, SubprocessError

    with pytest.raises(SubprocessError, match='Command "ls /dev/null /does/not/exist" returned exit status [0-9]+.\n>>> Standard Error\n') as error:
        raw_command(['ls', '/dev/null', '/does/not/exist'], True)

    assert '/dev/null' in error.value.stdout
    assert '/does/not/exist' in error.value.stderr


def test_failed_non_interactive_command() -> None:
    """Verify failed non-interactive non-captured commands raise a `SubprocessError` with `stdout` and `stderr` set to an empty string."""
    from ansible_test._internal.util import raw_command, SubprocessError

    with pytest.raises(SubprocessError, match='Command "ls /dev/null /does/not/exist" returned exit status [0-9]+.') as error:
        raw_command(['ls', '/dev/null', '/does/not/exist'], False)

    assert error.value.stdout == ''
    assert error.value.stderr == ''


def test_failed_interactive_command() -> None:
    """Verify failed interactive commands raise a `SubprocessError` with `stdout` and `stderr` set to `None`."""
    from ansible_test._internal.util import raw_command, SubprocessError

    with pytest.raises(SubprocessError, match='Command "ls /dev/null /does/not/exist" returned exit status [0-9]+.') as error:
        raw_command(['ls', '/dev/null', '/does/not/exist'], False, interactive=True)

    assert error.value.stdout is None
    assert error.value.stderr is None


@pytest.mark.parametrize("args,filters,expected", (
    (
        # args after a known option must be separated from existing args
        ['--color', '--docker', 'default', 'ping', 'split'],
        {'--docker': 1},
        ['--color', '--', 'ping', 'split'],
    ),
    (
        # args after '--' are not options
        ['--', '--color'],
        {'--color': 1},
        ['--', '--color'],
    ),
    (
        # args after a known option must be separated from existing args without conflicting with an explicit '--'
        ['--color', '--docker', 'default', 'ping', '--', 'split'],
        {'--docker': 1},
        ['--color', '--', 'ping', 'split'],
    ),
    (
        # args before options are properly handled
        ['ping', '--docker', 'default', '-v'],
        {'--docker': 1},
        ['ping', '-v'],
    ),
    (
        # no options
        ['ping', 'split'],
        {'--docker': 1},
        ['ping', 'split'],
    ),
    (
        # only options
        ['--color', '--docker', 'default', '-v'],
        {'--docker': 1},
        ['--color', '-v'],
    )
))
def test_filter_args(args: list[str], filters: dict[str, int], expected: list[str]) -> None:
    """Verify arg filtering properly handles various scenarios."""
    from ansible_test._internal.util import filter_args

    assert filter_args(args, filters) == expected



from __future__ import annotations

import io as _io

from yaml.resolver import Resolver

from ansible.module_utils._internal._datatag import AnsibleTagHelper
from ansible.module_utils.common.yaml import HAS_LIBYAML
from ansible._internal._datatag import _tags

from ._constructor import AnsibleConstructor, AnsibleInstrumentedConstructor

if HAS_LIBYAML:
    from yaml.cyaml import CParser

    class _Parser(CParser):
        def __init__(self, stream: str | bytes | _io.IOBase) -> None:
            if isinstance(stream, (str, bytes)):
                stream = AnsibleTagHelper.untag(stream)  # PyYAML + libyaml barfs on str/bytes subclasses

            CParser.__init__(self, stream)

else:
    from yaml.composer import Composer
    from yaml.reader import Reader
    from yaml.scanner import Scanner
    from yaml.parser import Parser

    class _Parser(Reader, Scanner, Parser, Composer):  # type: ignore[no-redef]
        def __init__(self, stream: str | bytes | _io.IOBase) -> None:
            Reader.__init__(self, stream)
            Scanner.__init__(self)
            Parser.__init__(self)
            Composer.__init__(self)


class _YamlParser(_Parser):
    def __init__(self, stream: str | bytes | _io.IOBase) -> None:
        super().__init__(stream)

        # The Python implementation of PyYAML (yaml.reader.Reader) provides self.name.
        # However, it will fall back to "<...>" in various cases.
        # The C implementation of PyYAML does not provide self.name.
        # To provide consistency, name retrieval is re-implemented here.
        self.name = getattr(stream, 'name', None)


class AnsibleInstrumentedLoader(_YamlParser, AnsibleInstrumentedConstructor, Resolver):
    """Ansible YAML loader which supports Ansible custom behavior such as `Origin` tagging, but no Ansible-specific YAML tags."""

    def __init__(self, stream: str | bytes | _io.IOBase) -> None:
        _YamlParser.__init__(self, stream)

        AnsibleInstrumentedConstructor.__init__(
            self,
            origin=_tags.Origin.get_or_create_tag(stream, self.name),
            trusted_as_template=_tags.TrustedAsTemplate.is_tagged_on(stream),
        )

        Resolver.__init__(self)


class AnsibleLoader(_YamlParser, AnsibleConstructor, Resolver):
    """Ansible loader which supports Ansible custom behavior such as `Origin` tagging, as well as Ansible-specific YAML tags."""

    def __init__(self, stream: str | bytes | _io.IOBase) -> None:
        _YamlParser.__init__(self, stream)

        AnsibleConstructor.__init__(
            self,
            origin=_tags.Origin.get_or_create_tag(stream, self.name),
            trusted_as_template=_tags.TrustedAsTemplate.is_tagged_on(stream),
        )

        Resolver.__init__(self)



# Copyright (c) 2022 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
    inventory: exercise_cache
    short_description: run tests against the specified cache plugin
    description:
      - This plugin doesn't modify inventory.
      - Load a cache plugin and test the inventory cache interface is dict-like.
      - Most inventory cache write methods only apply to the in-memory cache.
      - The 'flush' and 'set_cache' methods should be used to apply changes to the backing cache plugin.
      - The inventory cache read methods prefer the in-memory cache, and fall back to reading from the cache plugin.
    extends_documentation_fragment:
      - inventory_cache
    options:
      plugin:
        required: true
        description: name of the plugin (exercise_cache)
    cache_timeout:
        ini: []
        env: []
        cli: []
        default: 0  # never expire
"""

from ansible.errors import AnsibleError
from ansible.plugins.inventory import BaseInventoryPlugin, Cacheable
from ansible.utils.display import Display

from time import sleep


display = Display()


class InventoryModule(BaseInventoryPlugin, Cacheable):

    NAME = 'exercise_cache'

    test_cache_methods = [
        'test_plugin_name',
        'test_update_cache_if_changed',
        'test_set_cache',
        'test_load_whole_cache',
        'test_iter',
        'test_len',
        'test_get_missing_key',
        'test_get_expired_key',
        'test_initial_get',
        'test_get',
        'test_items',
        'test_keys',
        'test_values',
        'test_pop',
        'test_del',
        'test_set',
        'test_update',
        'test_flush',
    ]

    def verify_file(self, path):
        if not path.endswith(('exercise_cache.yml', 'exercise_cache.yaml',)):
            return False
        return super(InventoryModule, self).verify_file(path)

    def parse(self, inventory, loader, path, cache=None):
        super(InventoryModule, self).parse(inventory, loader, path)
        self._read_config_data(path)

        try:
            self.exercise_test_cache()
        except AnsibleError:
            raise
        except Exception as e:
            raise AnsibleError("Failed to run cache tests: {0}".format(e)) from e

    def exercise_test_cache(self):
        failed = []
        for test_name in self.test_cache_methods:
            try:
                getattr(self, test_name)()
            except AssertionError:
                failed.append(test_name)
            finally:
                self.cache.flush()
                self.cache.update_cache_if_changed()

        if failed:
            raise AnsibleError(f"Cache tests failed: {', '.join(failed)}")

    def test_equal(self, a, b):
        try:
            assert a == b
        except AssertionError:
            display.warning(f"Assertion {a} == {b} failed")
            raise

    def test_plugin_name(self):
        self.test_equal(self.cache._plugin_name, self.get_option('cache_plugin'))

    def test_update_cache_if_changed(self):
        self.cache._retrieved = {}
        self.cache._cache = {'foo': 'bar'}

        self.cache.update_cache_if_changed()

        self.test_equal(self.cache._retrieved, {'foo': 'bar'})
        self.test_equal(self.cache._cache, {'foo': 'bar'})

    def test_set_cache(self):
        cache_key1 = 'key1'
        cache1 = {'hosts': {'h1': {'foo': 'bar'}}}
        cache_key2 = 'key2'
        cache2 = {'hosts': {'h2': {}}}

        self.cache._cache = {cache_key1: cache1, cache_key2: cache2}
        self.cache.set_cache()

        self.test_equal(self.cache._plugin.contains(cache_key1), True)
        self.test_equal(self.cache._plugin.get(cache_key1), cache1)
        self.test_equal(self.cache._plugin.contains(cache_key2), True)
        self.test_equal(self.cache._plugin.get(cache_key2), cache2)

    def test_load_whole_cache(self):
        cache_data = {
            'key1': {'hosts': {'h1': {'foo': 'bar'}}},
            'key2': {'hosts': {'h2': {}}},
        }
        self.cache._cache = cache_data
        self.cache.set_cache()
        self.cache._cache = {}

        self.cache.load_whole_cache()
        self.test_equal(self.cache._cache, cache_data)

    def test_iter(self):
        cache_data = {
            'key1': {'hosts': {'h1': {'foo': 'bar'}}},
            'key2': {'hosts': {'h2': {}}},
        }
        self.cache._cache = cache_data
        self.test_equal(sorted(list(self.cache)), ['key1', 'key2'])

    def test_len(self):
        cache_data = {
            'key1': {'hosts': {'h1': {'foo': 'bar'}}},
            'key2': {'hosts': {'h2': {}}},
        }
        self.cache._cache = cache_data
        self.test_equal(len(self.cache), 2)

    def test_get_missing_key(self):
        # cache should behave like a dictionary
        # a missing key with __getitem__ should raise a KeyError
        try:
            self.cache['keyerror']
        except KeyError:
            pass
        else:
            assert False

        # get should return the default instead
        self.test_equal(self.cache.get('missing'), None)
        self.test_equal(self.cache.get('missing', 'default'), 'default')

    def _setup_expired(self):
        self.cache._cache = {'expired': True}
        self.cache.set_cache()

        # empty the in-memory info to test loading the key
        # keys that expire mid-use do not cause errors
        self.cache._cache = {}
        self.cache._retrieved = {}
        self.cache._plugin._cache = {}

        self.cache._plugin.set_option('timeout', 1)
        self.cache._plugin._timeout = 1
        sleep(2)

    def _cleanup_expired(self):
        # Set cache timeout back to never
        self.cache._plugin.set_option('timeout', 0)
        self.cache._plugin._timeout = 0

    def test_get_expired_key(self):
        if not hasattr(self.cache._plugin, '_timeout'):
            # DB-backed caches do not have a standard timeout interface
            return

        self._setup_expired()
        try:
            self.cache['expired']
        except KeyError:
            pass
        else:
            assert False
        finally:
            self._cleanup_expired()

        self._setup_expired()
        try:
            self.test_equal(self.cache.get('expired'), None)
            self.test_equal(self.cache.get('expired', 'default'), 'default')
        finally:
            self._cleanup_expired()

    def test_initial_get(self):
        # test cache behaves like a dictionary

        # set the cache to test getting a key that exists
        k1 = {'hosts': {'h1': {'foo': 'bar'}}}
        k2 = {'hosts': {'h2': {}}}
        self.cache._cache = {'key1': k1, 'key2': k2}
        self.cache.set_cache()

        # empty the in-memory info to test loading the key from the plugin
        self.cache._cache = {}
        self.cache._retrieved = {}
        self.cache._plugin._cache = {}

        self.test_equal(self.cache['key1'], k1)

        # empty the in-memory info to test loading the key from the plugin
        self.cache._cache = {}
        self.cache._retrieved = {}
        self.cache._plugin._cache = {}

        self.test_equal(self.cache.get('key1'), k1)

    def test_get(self):
        # test cache behaves like a dictionary

        # set the cache to test getting a key that exists
        k1 = {'hosts': {'h1': {'foo': 'bar'}}}
        k2 = {'hosts': {'h2': {}}}
        self.cache._cache = {'key1': k1, 'key2': k2}
        self.cache.set_cache()

        self.test_equal(self.cache['key1'], k1)
        self.test_equal(self.cache.get('key1'), k1)

    def test_items(self):
        self.test_equal(self.cache.items(), {}.items())

        test_items = {'hosts': {'host1': {'foo': 'bar'}}}
        self.cache._cache = test_items
        self.test_equal(self.cache.items(), test_items.items())

    def test_keys(self):
        self.test_equal(self.cache.keys(), {}.keys())

        test_items = {'hosts': {'host1': {'foo': 'bar'}}}
        self.cache._cache = test_items
        self.test_equal(self.cache.keys(), test_items.keys())

    def test_values(self):
        self.test_equal(list(self.cache.values()), list({}.values()))

        test_items = {'hosts': {'host1': {'foo': 'bar'}}}
        self.cache._cache = test_items
        self.test_equal(list(self.cache.values()), list(test_items.values()))

    def test_pop(self):
        try:
            self.cache.pop('missing')
        except KeyError:
            pass
        else:
            assert False

        self.test_equal(self.cache.pop('missing', 'default'), 'default')

        self.cache._cache = {'cache_key': 'cache'}
        self.test_equal(self.cache.pop('cache_key'), 'cache')

        # test backing plugin cache isn't modified
        cache_key1 = 'key1'
        cache1 = {'hosts': {'h1': {'foo': 'bar'}}}
        cache_key2 = 'key2'
        cache2 = {'hosts': {'h2': {}}}

        self.cache._cache = {cache_key1: cache1, cache_key2: cache2}
        self.cache.set_cache()

        self.test_equal(self.cache.pop('key1'), cache1)
        self.test_equal(self.cache._cache, {cache_key2: cache2})
        self.test_equal(self.cache._plugin._cache, {cache_key1: cache1, cache_key2: cache2})

    def test_del(self):
        try:
            del self.cache['missing']
        except KeyError:
            pass
        else:
            assert False

        cache_key1 = 'key1'
        cache1 = {'hosts': {'h1': {'foo': 'bar'}}}
        cache_key2 = 'key2'
        cache2 = {'hosts': {'h2': {}}}

        self.cache._cache = {cache_key1: cache1, cache_key2: cache2}
        self.cache.set_cache()

        del self.cache['key1']

        self.test_equal(self.cache._cache, {cache_key2: cache2})
        self.test_equal(self.cache._plugin._cache, {cache_key1: cache1, cache_key2: cache2})

    def test_set(self):
        cache_key = 'key1'
        hosts = {'hosts': {'h1': {'foo': 'bar'}}}
        self.cache[cache_key] = hosts

        self.test_equal(self.cache._cache, {cache_key: hosts})
        self.test_equal(self.cache._plugin._cache, {})

    def test_update(self):
        cache_key1 = 'key1'
        cache1 = {'hosts': {'h1': {'foo': 'bar'}}}
        cache_key2 = 'key2'
        cache2 = {'hosts': {'h2': {}}}

        self.cache._cache = {cache_key1: cache1}
        self.cache.update({cache_key2: cache2})
        self.test_equal(self.cache._cache, {cache_key1: cache1, cache_key2: cache2})

    def test_flush(self):
        cache_key1 = 'key1'
        cache1 = {'hosts': {'h1': {'foo': 'bar'}}}
        cache_key2 = 'key2'
        cache2 = {'hosts': {'h2': {}}}

        self.cache._cache = {cache_key1: cache1, cache_key2: cache2}
        self.cache.set_cache()

        # Unlike the dict write methods, cache.flush() flushes the backing plugin
        self.cache.flush()

        self.test_equal(self.cache._cache, {})
        self.test_equal(self.cache._plugin._cache, {})



"""
Backwards compatibility profile for serialization other than inventory (which should use inventory_legacy for backward-compatible trust behavior).
Behavior is equivalent to pre 2.19 `AnsibleJSONEncoder` with vault_to_text=True.
"""

from __future__ import annotations as _annotations

import datetime as _datetime
import typing as _t

from ansible._internal import _json
from ansible._internal._datatag import _tags
from ansible.module_utils._internal import _datatag
from ansible.module_utils._internal._json import _profiles
from ansible.parsing import vault as _vault


class _Untrusted:
    """
    Temporarily wraps strings which are not trusted for templating.
    Used before serialization of strings not tagged TrustedAsTemplate when trust inversion is enabled and trust is allowed in the string's context.
    Used during deserialization of `__ansible_unsafe` strings to indicate they should not be tagged TrustedAsTemplate.
    """

    __slots__ = ('value',)

    def __init__(self, value: str) -> None:
        self.value = value


class _LegacyVariableVisitor(_json.AnsibleVariableVisitor):
    """Variable visitor that supports optional trust inversion for legacy serialization."""

    def __init__(
        self,
        *,
        trusted_as_template: bool = False,
        invert_trust: bool = False,
        origin: _tags.Origin | None = None,
        convert_mapping_to_dict: bool = False,
        convert_sequence_to_list: bool = False,
        convert_custom_scalars: bool = False,
    ):
        super().__init__(
            trusted_as_template=trusted_as_template,
            origin=origin,
            convert_mapping_to_dict=convert_mapping_to_dict,
            convert_sequence_to_list=convert_sequence_to_list,
            convert_custom_scalars=convert_custom_scalars,
            encrypted_string_behavior=_json.EncryptedStringBehavior.PRESERVE,
        )

        self.invert_trust = invert_trust

        if trusted_as_template and invert_trust:
            raise ValueError('trusted_as_template is mutually exclusive with invert_trust')

    @property
    def _allow_trust(self) -> bool:
        """
        This profile supports trust application in all contexts.
        Derived implementations can override this behavior for application-dependent/schema-aware trust.
        """
        return True

    def _early_visit(self, value, value_type) -> _t.Any:
        """Similar to base implementation, but supports an intermediate wrapper for trust inversion."""
        if value_type in (str, _datatag._AnsibleTaggedStr):
            # apply compatibility behavior
            if self.trusted_as_template and self._allow_trust:
                result = _tags.TrustedAsTemplate().tag(value)
            elif self.invert_trust and not _tags.TrustedAsTemplate.is_tagged_on(value) and self._allow_trust:
                result = _Untrusted(value)
            else:
                result = value
        elif value_type is _Untrusted:
            result = value.value
        else:
            result = _json._sentinel

        return result


class _Profile(_profiles._JSONSerializationProfile["Encoder", "Decoder"]):
    visitor_type = _LegacyVariableVisitor

    @classmethod
    def serialize_untrusted(cls, value: _Untrusted) -> dict[str, str] | str:
        return dict(
            __ansible_unsafe=_datatag.AnsibleTagHelper.untag(value.value),
        )

    @classmethod
    def serialize_tagged_str(cls, value: _datatag.AnsibleTaggedObject) -> _t.Any:
        if ciphertext := _vault.VaultHelper.get_ciphertext(value, with_tags=False):
            return dict(
                __ansible_vault=ciphertext,
            )

        return _datatag.AnsibleTagHelper.untag(value)

    @classmethod
    def deserialize_unsafe(cls, value: dict[str, _t.Any]) -> _Untrusted:
        ansible_unsafe = value['__ansible_unsafe']

        if type(ansible_unsafe) is not str:  # pylint: disable=unidiomatic-typecheck
            raise TypeError(f"__ansible_unsafe is {type(ansible_unsafe)} not {str}")

        return _Untrusted(ansible_unsafe)

    @classmethod
    def deserialize_vault(cls, value: dict[str, _t.Any]) -> _vault.EncryptedString:
        ansible_vault = value['__ansible_vault']

        if type(ansible_vault) is not str:  # pylint: disable=unidiomatic-typecheck
            raise TypeError(f"__ansible_vault is {type(ansible_vault)} not {str}")

        encrypted_string = _vault.EncryptedString(ciphertext=ansible_vault)

        return encrypted_string

    @classmethod
    def serialize_encrypted_string(cls, value: _vault.EncryptedString) -> dict[str, str]:
        return dict(
            __ansible_vault=_vault.VaultHelper.get_ciphertext(value, with_tags=False),
        )

    @classmethod
    def post_init(cls) -> None:
        cls.serialize_map = {
            set: cls.serialize_as_list,
            tuple: cls.serialize_as_list,
            _datetime.date: cls.serialize_as_isoformat,  # existing devel behavior
            _datetime.time: cls.serialize_as_isoformat,  # always failed pre-2.18, so okay to include for consistency
            _datetime.datetime: cls.serialize_as_isoformat,  # existing devel behavior
            _datatag._AnsibleTaggedDate: cls.discard_tags,
            _datatag._AnsibleTaggedTime: cls.discard_tags,
            _datatag._AnsibleTaggedDateTime: cls.discard_tags,
            _vault.EncryptedString: cls.serialize_encrypted_string,
            _datatag._AnsibleTaggedStr: cls.serialize_tagged_str,  # for VaultedValue tagged str
            _datatag._AnsibleTaggedInt: cls.discard_tags,
            _datatag._AnsibleTaggedFloat: cls.discard_tags,
            _datatag._AnsibleTaggedList: cls.discard_tags,
            _datatag._AnsibleTaggedSet: cls.discard_tags,
            _datatag._AnsibleTaggedTuple: cls.discard_tags,
            _datatag._AnsibleTaggedDict: cls.discard_tags,
            _Untrusted: cls.serialize_untrusted,  # equivalent to AnsibleJSONEncoder(preprocess_unsafe=True) in devel
        }

        cls.deserialize_map = {
            '__ansible_unsafe': cls.deserialize_unsafe,
            '__ansible_vault': cls.deserialize_vault,
        }

        cls.handle_key = cls._handle_key_str_fallback  # type: ignore[method-assign]  # legacy stdlib-compatible key behavior

    @classmethod
    def pre_serialize(cls, encoder: Encoder, o: _t.Any) -> _t.Any:
        # DTFIX7: these conversion args probably aren't needed
        avv = cls.visitor_type(invert_trust=True, convert_mapping_to_dict=True, convert_sequence_to_list=True, convert_custom_scalars=True)

        return avv.visit(o)

    @classmethod
    def post_deserialize(cls, decoder: Decoder, o: _t.Any) -> _t.Any:
        avv = cls.visitor_type(trusted_as_template=decoder._trusted_as_template, origin=decoder._origin)

        return avv.visit(o)


class Encoder(_profiles.AnsibleProfileJSONEncoder):
    _profile = _Profile


class Decoder(_profiles.AnsibleProfileJSONDecoder):
    _profile = _Profile

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        # NB: these can only be sampled properly when loading strings, eg, `json.loads`; the global `json.load` function does not expose the file-like to us
        self._origin: _tags.Origin | None = None
        self._trusted_as_template: bool = False

    def raw_decode(self, s: str, idx: int = 0) -> tuple[_t.Any, int]:
        self._origin = _tags.Origin.get_tag(s)
        self._trusted_as_template = _tags.TrustedAsTemplate.is_tagged_on(s)

        return super().raw_decode(s, idx)



# -*- coding: utf-8 -*-
# Copyright:
#   (c) 2018 Ansible Project
# License: GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.modules.copy import AnsibleModuleError, split_pre_existing_dir

from ansible.module_utils.basic import AnsibleModule

THREE_DIRS_DATA: tuple[tuple[str, tuple[str, list[str]] | None, tuple[str, list[str]], tuple[str, list[str]], tuple[str, list[str]]], ...] = (
    ('/dir1/dir2',
     # 0 existing dirs: error (because / should always exist)
     None,
     # 1 existing dir:
     ('/', ['dir1', 'dir2']),
     # 2 existing dirs:
     ('/dir1', ['dir2']),
     # 3 existing dirs:
     ('/dir1/dir2', [])
     ),
    ('/dir1/dir2/',
     # 0 existing dirs: error (because / should always exist)
     None,
     # 1 existing dir:
     ('/', ['dir1', 'dir2']),
     # 2 existing dirs:
     ('/dir1', ['dir2']),
     # 3 existing dirs:
     ('/dir1/dir2', [])
     ),
)


TWO_DIRS_DATA: tuple[tuple[str, tuple[str, list[str]] | None, tuple[str, list[str]], tuple[str, list[str]]], ...] = (
    ('dir1/dir2',
     # 0 existing dirs:
     ('.', ['dir1', 'dir2']),
     # 1 existing dir:
     ('dir1', ['dir2']),
     # 2 existing dirs:
     ('dir1/dir2', []),
     # 3 existing dirs: Same as 2 because we never get to the third
     ),
    ('dir1/dir2/',
     # 0 existing dirs:
     ('.', ['dir1', 'dir2']),
     # 1 existing dir:
     ('dir1', ['dir2']),
     # 2 existing dirs:
     ('dir1/dir2', []),
     # 3 existing dirs: Same as 2 because we never get to the third
     ),
    ('/dir1',
     # 0 existing dirs: error (because / should always exist)
     None,
     # 1 existing dir:
     ('/', ['dir1']),
     # 2 existing dirs:
     ('/dir1', []),
     # 3 existing dirs: Same as 2 because we never get to the third
     ),
    ('/dir1/',
     # 0 existing dirs: error (because / should always exist)
     None,
     # 1 existing dir:
     ('/', ['dir1']),
     # 2 existing dirs:
     ('/dir1', []),
     # 3 existing dirs: Same as 2 because we never get to the third
     ),
)
TWO_DIRS_DATA += tuple(item[:4] for item in THREE_DIRS_DATA)


ONE_DIR_DATA: tuple[tuple[str, tuple[str, list[str]] | None, tuple[str, list[str]]], ...] = (
    ('dir1',
     # 0 existing dirs:
     ('.', ['dir1']),
     # 1 existing dir:
     ('dir1', []),
     # 2 existing dirs: Same as 1 because we never get to the third
     ),
    ('dir1/',
     # 0 existing dirs:
     ('.', ['dir1']),
     # 1 existing dir:
     ('dir1', []),
     # 2 existing dirs: Same as 1 because we never get to the third
     ),
)
ONE_DIR_DATA += tuple(item[:3] for item in TWO_DIRS_DATA)


@pytest.mark.parametrize('directory, expected', ((d[0], d[4]) for d in THREE_DIRS_DATA))
@pytest.mark.xfail(reason='broken test and/or code, original test missing assert', strict=False)
def test_split_pre_existing_dir_three_levels_exist(directory, expected, mocker):
    mocker.patch('os.path.exists', side_effect=[True, True, True])
    assert split_pre_existing_dir(directory) == expected


@pytest.mark.parametrize('directory, expected', ((d[0], d[3]) for d in TWO_DIRS_DATA))
@pytest.mark.xfail(reason='broken test and/or code, original test missing assert', strict=False)
def test_split_pre_existing_dir_two_levels_exist(directory, expected, mocker):
    mocker.patch('os.path.exists', side_effect=[True, True, False])
    assert split_pre_existing_dir(directory) == expected


@pytest.mark.parametrize('directory, expected', ((d[0], d[2]) for d in ONE_DIR_DATA))
@pytest.mark.xfail(reason='broken test and/or code, original test missing assert', strict=False)
def test_split_pre_existing_dir_one_level_exists(directory, expected, mocker):
    mocker.patch('os.path.exists', side_effect=[True, False, False])
    assert split_pre_existing_dir(directory) == expected


@pytest.mark.parametrize('directory', (d[0] for d in ONE_DIR_DATA if d[1] is None))
def test_split_pre_existing_dir_root_does_not_exist(directory, mocker):
    mocker.patch('os.path.exists', return_value=False)
    with pytest.raises(AnsibleModuleError) as excinfo:
        split_pre_existing_dir(directory)
    assert excinfo.value.results['msg'].startswith("The '/' directory doesn't exist on this machine.")


@pytest.mark.parametrize('directory, expected', ((d[0], d[1]) for d in ONE_DIR_DATA if not d[0].startswith('/')))
@pytest.mark.xfail(reason='broken test and/or code, original test missing assert', strict=False)
def test_split_pre_existing_dir_working_dir_exists(directory, expected, mocker):
    mocker.patch('os.path.exists', return_value=False)
    assert split_pre_existing_dir(directory) == expected


#
# Info helpful for making new test cases:
#
# base_mode = {
# 'dir no perms':   0o040000,
# 'file no perms':  0o100000,
# 'dir all perms':  0o040000 | 0o777,
# 'file all perms': 0o100000 | 0o777}
#
# perm_bits = {
# 'x': 0b001,
# 'w': 0b010,
# 'r': 0b100}
#
# role_shift = {
# 'u': 6,
# 'g': 3,
# 'o': 0}

DATA = (  # Going from no permissions to setting all for user, group, and/or other
    (0o040000, u'a+rwx', 0o0777),
    (0o040000, u'u+rwx,g+rwx,o+rwx', 0o0777),
    (0o040000, u'o+rwx', 0o0007),
    (0o040000, u'g+rwx', 0o0070),
    (0o040000, u'u+rwx', 0o0700),

    # Going from all permissions to none for user, group, and/or other
    (0o040777, u'a-rwx', 0o0000),
    (0o040777, u'u-rwx,g-rwx,o-rwx', 0o0000),
    (0o040777, u'o-rwx', 0o0770),
    (0o040777, u'g-rwx', 0o0707),
    (0o040777, u'u-rwx', 0o0077),

    # now using absolute assignment from None to a set of perms
    (0o040000, u'a=rwx', 0o0777),
    (0o040000, u'u=rwx,g=rwx,o=rwx', 0o0777),
    (0o040000, u'o=rwx', 0o0007),
    (0o040000, u'g=rwx', 0o0070),
    (0o040000, u'u=rwx', 0o0700),

    # X effect on files and dirs
    (0o040000, u'a+X', 0o0111),
    (0o100000, u'a+X', 0),
    (0o040000, u'a=X', 0o0111),
    (0o100000, u'a=X', 0),
    (0o040777, u'a-X', 0o0666),
    # Same as chmod but is it a bug?
    # chmod a-X statfile <== removes execute from statfile
    (0o100777, u'a-X', 0o0666),

    # Verify X uses computed not original mode
    (0o100777, u'a=,u=rX', 0o0400),
    (0o040777, u'a=,u=rX', 0o0500),

    # Multiple permissions
    (0o040000, u'u=rw-x+X,g=r-x+X,o=r-x+X', 0o0755),
    (0o100000, u'u=rw-x+X,g=r-x+X,o=r-x+X', 0o0644),
)

UMASK_DATA = (
    (0o100000, '+rwx', 0o770),
    (0o100777, '-rwx', 0o007),
)

INVALID_DATA = (
    (0o040000, u'a=foo', "bad symbolic permission for mode: a=foo"),
    (0o040000, u'f=rwx', "bad symbolic permission for mode: f=rwx"),
    (0o100777, u'of=r', "bad symbolic permission for mode: of=r"),

    (0o100777, u'ao=r', "bad symbolic permission for mode: ao=r"),
    (0o100777, u'oa=r', "bad symbolic permission for mode: oa=r"),
)


@pytest.mark.parametrize('stat_info, mode_string, expected', DATA)
def test_good_symbolic_modes(mocker, stat_info, mode_string, expected):
    mock_stat = mocker.MagicMock()
    mock_stat.st_mode = stat_info
    assert AnsibleModule._symbolic_mode_to_octal(mock_stat, mode_string) == expected


@pytest.mark.parametrize('stat_info, mode_string, expected', UMASK_DATA)
def test_umask_with_symbolic_modes(mocker, stat_info, mode_string, expected):
    mock_umask = mocker.patch('os.umask')
    mock_umask.return_value = 0o7

    mock_stat = mocker.MagicMock()
    mock_stat.st_mode = stat_info

    assert AnsibleModule._symbolic_mode_to_octal(mock_stat, mode_string) == expected


@pytest.mark.parametrize('stat_info, mode_string, expected', INVALID_DATA)
def test_invalid_symbolic_modes(mocker, stat_info, mode_string, expected):
    mock_stat = mocker.MagicMock()
    mock_stat.st_mode = stat_info
    with pytest.raises(ValueError) as exc:
        assert AnsibleModule._symbolic_mode_to_octal(mock_stat, mode_string) == 'blah'
    assert exc.match(expected)



from __future__ import annotations

import ast

from ansible.errors import AnsibleTypeError


def python_literal_eval(value: object, ignore_errors=False) -> object:
    try:
        if isinstance(value, str):
            return ast.literal_eval(value)

        raise AnsibleTypeError("The `value` to eval must be a string.", obj=value)
    except Exception:
        if ignore_errors:
            return value

        raise


class FilterModule(object):
    @staticmethod
    def filters():
        return dict(python_literal_eval=python_literal_eval)



# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import ast

import yaml

from ansible import constants as C
from ansible.errors import AnsibleError, AnsibleParserError
from ansible.module_utils.common.text.converters import to_text, to_native
from ansible.parsing.yaml.loader import AnsibleLoader
from ansible.utils.display import Display
from ansible._internal._datatag import _tags

display = Display()


string_to_vars = {
    'DOCUMENTATION': 'doc',
    'EXAMPLES': 'plainexamples',
    'RETURN': 'returndocs',
    'ANSIBLE_METADATA': 'metadata',  # NOTE: now unused, but kept for backwards compat
}


def _init_doc_dict():
    """ initialize a return dict for docs with the expected structure """
    return {k: None for k in string_to_vars.values()}


def read_docstring_from_yaml_file(filename, verbose=True, ignore_errors=True):
    """ Read docs from 'sidecar' yaml file doc for a plugin """

    data = _init_doc_dict()
    file_data = {}

    try:
        with open(filename, 'rb') as yamlfile:
            file_data = yaml.load(yamlfile, Loader=AnsibleLoader)
    except Exception as ex:
        msg = f"Unable to parse yaml file {filename}"
        # DTFIX-FUTURE: find a better pattern for this (can we use the new optional error behavior?)
        if not ignore_errors:
            raise AnsibleParserError(f'{msg}.') from ex
        elif verbose:
            display.error(f'{msg}: {ex}')

    if file_data:
        for key in string_to_vars:
            data[string_to_vars[key]] = file_data.get(key, None)

    return data


def read_docstring_from_python_file(filename, verbose=True, ignore_errors=True):
    """
    Use ast to search for assignment of the DOCUMENTATION and EXAMPLES variables in the given file.
    Parse DOCUMENTATION from YAML and return the YAML doc or None together with EXAMPLES, as plain text.
    """
    data = _init_doc_dict()

    try:
        with open(filename, 'rb') as b_module_data:
            M = ast.parse(b_module_data.read())

        for child in M.body:
            if isinstance(child, ast.Assign):
                for t in child.targets:
                    try:
                        theid = t.id
                    except AttributeError:
                        # skip errors can happen when trying to use the normal code
                        display.warning("Building documentation, failed to assign id for %s on %s, skipping" % (t, filename))
                        continue

                    if theid in string_to_vars:
                        varkey = string_to_vars[theid]
                        if isinstance(child.value, ast.Dict):
                            data[varkey] = ast.literal_eval(child.value)
                        else:
                            if theid == 'EXAMPLES':
                                # examples 'can' be yaml, but even if so, we dont want to parse as such here
                                # as it can create undesired 'objects' that don't display well as docs.
                                data[varkey] = to_text(child.value.value)
                            else:
                                # string should be yaml if already not a dict
                                child_value = _tags.Origin(path=filename, line_num=child.value.lineno).tag(child.value.value)
                                data[varkey] = yaml.load(child_value, Loader=AnsibleLoader)

                        display.debug('Documentation assigned: %s' % varkey)

    except Exception as ex:
        msg = f"Unable to parse documentation in python file {filename!r}"
        # DTFIX-FUTURE: better pattern to conditionally raise/display
        if not ignore_errors:
            raise AnsibleParserError(f'{msg}.') from ex
        elif verbose:
            display.error(f'{msg}: {ex}.')

    return data


def read_docstring(filename, verbose=True, ignore_errors=True):
    """ returns a documentation dictionary from Ansible plugin docstrings """

    # NOTE: adjacency of doc file to code file is responsibility of caller
    if filename.endswith(C.YAML_DOC_EXTENSIONS):
        docstring = read_docstring_from_yaml_file(filename, verbose=verbose, ignore_errors=ignore_errors)
    elif filename.endswith(C.PYTHON_DOC_EXTENSIONS):
        docstring = read_docstring_from_python_file(filename, verbose=verbose, ignore_errors=ignore_errors)
    elif not ignore_errors:
        raise AnsibleError("Unknown documentation format: %s" % to_native(filename))

    if not docstring and not ignore_errors:
        raise AnsibleError("Unable to parse documentation for: %s" % to_native(filename))

    # cause seealso is specially processed from 'doc' later on
    # TODO: stop any other 'overloaded' implementation in main doc
    docstring['seealso'] = None

    return docstring


def read_docstub(filename):
    """
    Quickly find short_description using string methods instead of node parsing.
    This does not return a full set of documentation strings and is intended for
    operations like ansible-doc -l.
    """

    in_documentation = False
    capturing = False
    indent_detection = ''
    doc_stub = []

    with open(filename, 'r') as t_module_data:
        for line in t_module_data:
            if in_documentation:
                # start capturing the stub until indentation returns
                if capturing and line.startswith(indent_detection):
                    doc_stub.append(line)

                elif capturing and not line.startswith(indent_detection):
                    break

                elif line.lstrip().startswith('short_description:'):
                    capturing = True
                    # Detect that the short_description continues on the next line if it's indented more
                    # than short_description itself.
                    indent_detection = ' ' * (len(line) - len(line.lstrip()) + 1)
                    doc_stub.append(line)

            elif line.startswith('DOCUMENTATION') and ('=' in line or ':' in line):
                in_documentation = True

    short_description = r''.join(doc_stub).strip().rstrip('.')
    data = yaml.load(_tags.Origin(path=str(filename)).tag(short_description), Loader=AnsibleLoader)

    return data



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.module_utils.common.text.converters import to_native
from ansible.module_utils.common.validation import check_type_str, _check_type_str_no_conversion


TEST_CASES = (
    ('string', 'string'),
    (None, '',),  # 2.19+ relaxed restriction on None<->empty for backward compatibility
    (100, '100'),
    (1.5, '1.5'),
    ({'k1': 'v1'}, "{'k1': 'v1'}"),
    ([1, 2, 'three'], "[1, 2, 'three']"),
    ((1, 2,), '(1, 2)'),
)


@pytest.mark.parametrize('value, expected', TEST_CASES)
def test_check_type_str(value, expected):
    assert expected == check_type_str(value)


@pytest.mark.parametrize('value, expected', TEST_CASES[2:])
def test_check_type_str_no_conversion(value, expected):
    with pytest.raises(TypeError) as e:
        _check_type_str_no_conversion(value)
    assert 'is not a string and conversion is not allowed' in to_native(e.value)



"""Data tagging aware wire format for controller to module communication."""

from __future__ import annotations as _annotations

import datetime as _datetime

from ... import _datatag
from .. import _profiles


class _Profile(_profiles._JSONSerializationProfile["Encoder", "Decoder"]):
    encode_strings_as_utf8 = True

    @classmethod
    def post_init(cls) -> None:
        cls.serialize_map = {}
        cls.serialize_map.update(cls._common_discard_tags)
        cls.serialize_map.update(
            {
                # The bytes type is not supported, use str instead (future module profiles may support a bytes wrapper distinct from `bytes`).
                set: cls.serialize_as_list,  # legacy _json_encode_fallback behavior
                tuple: cls.serialize_as_list,  # JSONEncoder built-in behavior
                _datetime.date: _datatag.AnsibleSerializableDate,
                _datetime.time: _datatag.AnsibleSerializableTime,
                _datetime.datetime: _datatag.AnsibleSerializableDateTime,
            }
        )


class Encoder(_profiles.AnsibleProfileJSONEncoder):
    _profile = _Profile


class Decoder(_profiles.AnsibleProfileJSONDecoder):
    _profile = _Profile



from __future__ import annotations

import dataclasses
import typing as t

from ansible.module_utils._internal import _datatag, _messages


@dataclasses.dataclass(**_datatag._tag_dataclass_kwargs)
class Deprecated(_datatag.AnsibleDatatagBase):
    msg: str
    help_text: t.Optional[str] = None
    date: t.Optional[str] = None
    version: t.Optional[str] = None
    deprecator: t.Optional[_messages.PluginInfo] = None
    formatted_traceback: t.Optional[str] = None



from __future__ import annotations

import ast
import collections.abc as c
import dataclasses
import enum
import pathlib
import tempfile
import types
import typing as t

from collections import ChainMap

import jinja2.nodes

from jinja2 import pass_context, defaults, TemplateSyntaxError, FileSystemLoader
from jinja2.environment import Environment, Template, TemplateModule, TemplateExpression
from jinja2.compiler import Frame
from jinja2.lexer import TOKEN_VARIABLE_BEGIN, TOKEN_VARIABLE_END, TOKEN_STRING, Lexer
from jinja2.nativetypes import NativeCodeGenerator
from jinja2.nodes import Const, EvalContext
from jinja2.runtime import Context, Macro
from jinja2.sandbox import SandboxedEnvironment
from jinja2.utils import missing, LRUCache

from ansible.utils.display import Display
from ansible.errors import AnsibleVariableTypeError, AnsibleTemplateSyntaxError, AnsibleTemplateError
from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils._internal._datatag import (
    _AnsibleTaggedDict,
    _AnsibleTaggedList,
    _AnsibleTaggedTuple,
    _AnsibleTaggedStr,
    AnsibleTagHelper,
)

from ansible._internal._errors._handler import ErrorAction
from ansible._internal._datatag._tags import Origin, TrustedAsTemplate

from ._access import AnsibleAccessContext
from ._datatag import _JinjaConstTemplate
from ._utils import LazyOptions
from ._jinja_common import (
    MarkerError,
    Marker,
    CapturedExceptionMarker,
    UndefinedMarker,
    _TemplateConfig,
    TruncationMarker,
    validate_arg_type,
    JinjaCallContext,
    _SandboxMode,
)
from ._jinja_plugins import JinjaPluginIntercept, _query, _lookup, _now, _wrap_plugin_output, get_first_marker_arg, _DirectCall, _jinja_const_template_warning
from ._lazy_containers import (
    _AnsibleLazyTemplateMixin,
    _AnsibleLazyTemplateDict,
    _AnsibleLazyTemplateList,
    _AnsibleLazyAccessTuple,
    lazify_container_args,
    lazify_container_kwargs,
    lazify_container,
    register_known_types,
)
from ._utils import Omit, TemplateContext, PASS_THROUGH_SCALAR_VAR_TYPES

from ansible.module_utils._internal._json._profiles import _json_subclassable_scalar_types
from ansible.module_utils import _internal
from ansible.module_utils._internal import _ambient_context, _dataclass_validation
from ansible.plugins.loader import filter_loader, test_loader
from ansible.vars.hostvars import HostVars, HostVarsVars
from ...module_utils.datatag import native_type_name

JINJA2_OVERRIDE = '#jinja2:'
"""
String values prefixed with this sequence are interpreted as templates, even without template delimiters.
The values following this prefix up to the first newline are parsed as Jinja2 template overrides.
To include this literal value at the start of a string, a space or other character must precede it.
"""

JINJA_KEYWORDS = frozenset(
    {
        # scalar singletons (see jinja2.nodes.Name.can_assign)
        'true',
        'false',
        'none',
        'True',
        'False',
        'None',
        # other
        'not',  # unary operator always applicable to names
    }
)
"""Names which have special meaning to Jinja and cannot be resolved as variable names."""

display = Display()


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class TemplateOverrides:
    DEFAULT: t.ClassVar[t.Self]

    block_start_string: str = defaults.BLOCK_START_STRING
    block_end_string: str = defaults.BLOCK_END_STRING
    variable_start_string: str = defaults.VARIABLE_START_STRING
    variable_end_string: str = defaults.VARIABLE_END_STRING
    comment_start_string: str = defaults.COMMENT_START_STRING
    comment_end_string: str = defaults.COMMENT_END_STRING
    line_statement_prefix: str | None = defaults.LINE_STATEMENT_PREFIX
    line_comment_prefix: str | None = defaults.LINE_COMMENT_PREFIX
    trim_blocks: bool = True  # AnsibleEnvironment overrides this default, so don't use the Jinja default here
    lstrip_blocks: bool = defaults.LSTRIP_BLOCKS
    newline_sequence: t.Literal['\n', '\r\n', '\r'] = defaults.NEWLINE_SEQUENCE
    keep_trailing_newline: bool = defaults.KEEP_TRAILING_NEWLINE

    def __post_init__(self) -> None:
        pass  # overridden by _dataclass_validation._inject_post_init_validation

    def _post_validate(self) -> None:
        if not (self.block_start_string != self.variable_start_string != self.comment_start_string != self.block_start_string):
            raise ValueError('Block, variable and comment start strings must be different.')

    def overlay_kwargs(self) -> dict[str, t.Any]:
        """
        Return a dictionary of arguments for passing to Environment.overlay.
        The dictionary will be empty if all fields have their default value.
        """
        # DTFIX-FUTURE: calculate default/non-default during __post_init__
        fields = [(field, getattr(self, field.name)) for field in dataclasses.fields(self)]
        kwargs = {field.name: value for field, value in fields if value != field.default}

        return kwargs

    def _contains_start_string(self, value: str) -> bool:
        """Returns True if the given value contains a variable, block or comment start string."""
        # DTFIX-FUTURE: this is inefficient, use a compiled regex instead

        for marker in (self.block_start_string, self.variable_start_string, self.comment_start_string):
            if marker in value:
                return True

        return False

    def _starts_and_ends_with_jinja_delimiters(self, value: str) -> bool:
        """Returns True if the given value starts and ends with Jinja variable, block or comment delimiters."""
        # DTFIX-FUTURE: this is inefficient, use a compiled regex instead

        for marker in (self.block_start_string, self.variable_start_string, self.comment_start_string):
            if value.startswith(marker):
                break
        else:
            return False

        for marker in (self.block_end_string, self.variable_end_string, self.comment_end_string):
            if value.endswith(marker):
                return True

        return False

    def _extract_template_overrides(self, template: str) -> tuple[str, TemplateOverrides]:
        if template.startswith(JINJA2_OVERRIDE):
            eol = template.find('\n')

            if eol == -1:
                raise ValueError(f"Missing newline after {JINJA2_OVERRIDE!r} override.")

            line = template[len(JINJA2_OVERRIDE) : eol]
            template = template[eol + 1 :]
            override_kwargs = {}

            for pair in line.split(','):
                if not pair.strip():
                    raise ValueError(f"Empty {JINJA2_OVERRIDE!r} override pair not allowed.")

                if ':' not in pair:
                    raise ValueError(f"Missing key-value separator `:` in {JINJA2_OVERRIDE!r} override pair {pair!r}.")

                key, val = pair.split(':', 1)
                key = key.strip()

                if key not in _TEMPLATE_OVERRIDE_FIELD_NAMES:
                    raise ValueError(f"Invalid {JINJA2_OVERRIDE!r} override key {key!r}.")

                override_kwargs[key] = ast.literal_eval(val)

            overrides = dataclasses.replace(self, **override_kwargs)
        else:
            overrides = self

        return template, overrides

    def merge(self, kwargs: dict[str, t.Any] | None, /) -> TemplateOverrides:
        """Return a new instance based on the current instance with the given kwargs overridden."""
        if kwargs:
            return self.from_kwargs(dataclasses.asdict(self) | kwargs)

        return self

    @classmethod
    def from_kwargs(cls, kwargs: dict[str, t.Any] | None, /) -> TemplateOverrides:
        """TemplateOverrides instance factory; instances resolving to all default values will instead return the DEFAULT singleton for optimization."""
        if kwargs:
            value = cls(**kwargs)

            if value.overlay_kwargs():
                return value

        return cls.DEFAULT


_dataclass_validation.inject_post_init_validation(TemplateOverrides, allow_subclasses=True)

TemplateOverrides.DEFAULT = TemplateOverrides()

_TEMPLATE_OVERRIDE_FIELD_NAMES: t.Final[tuple[str, ...]] = tuple(sorted(field.name for field in dataclasses.fields(TemplateOverrides)))


class AnsibleContext(Context):
    """
    A custom context which intercepts resolve_or_missing() calls and
    runs them through AnsibleAccessContext. This allows usage of variables
    to be tracked. If needed, values can also be modified before being returned.
    """

    environment: AnsibleEnvironment  # narrow the type specified by the base

    def __init__(self, *args, **kwargs):
        super(AnsibleContext, self).__init__(*args, **kwargs)

    __repr__ = object.__repr__  # prevent Jinja from dumping vars in case this gets repr'd

    def get_all(self):
        """
        Override Jinja's default get_all to return all vars in the context as a ChainMap with a mutable layer at the bottom.
        This provides some isolation against accidental changes to inherited variable contexts without requiring copies.
        """
        layers = []

        if self.vars:
            layers.append(self.vars)
        if self.parent:
            layers.append(self.parent)

        # HACK: always include a sacrificial plain-dict on the bottom layer, since Jinja's debug and stacktrace rewrite code invokes
        # `__setitem__` outside a call context; this will ensure that it always occurs on a plain dict instead of a lazy one.
        return ChainMap({}, *layers)

    # noinspection PyShadowingBuiltins
    def derived(self, locals: t.Optional[t.Dict[str, t.Any]] = None) -> Context:
        # this is a clone of Jinja's impl of derived, but using our lazy-aware _new_context

        context = _new_context(
            environment=self.environment,
            template_name=self.name,
            blocks={},
            shared=True,
            jinja_locals=locals,
            jinja_vars=self.get_all(),
        )
        context.eval_ctx = self.eval_ctx
        context.blocks.update((k, list(v)) for k, v in self.blocks.items())
        return context

    def keys(self, *args, **kwargs):
        """Base Context delegates to `dict.keys` against `get_all`, which would fail since we return a ChainMap. No known usage."""
        raise NotImplementedError()

    def values(self, *args, **kwargs):
        """Base Context delegates to `dict.values` against `get_all`, which would fail since we return a ChainMap. No known usage."""
        raise NotImplementedError()

    def items(self, *args, **kwargs):
        """Base Context delegates to built-in `dict.items` against `get_all`, which would fail since we return a ChainMap. No known usage."""
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class ArgSmuggler:
    """
    Utility wrapper to wrap/unwrap args passed to Jinja `Template.render` and `TemplateExpression.__call__`.
    e.g., see https://github.com/pallets/jinja/blob/3.1.3/src/jinja2/environment.py#L1296 and
    https://github.com/pallets/jinja/blob/3.1.3/src/jinja2/environment.py#L1566.
    """

    jinja_vars: c.Mapping[str, t.Any] | None

    @classmethod
    def package_jinja_vars(cls, jinja_vars: c.Mapping[str, t.Any]) -> dict[str, ArgSmuggler]:
        """Wrap the supplied vars dict in an ArgSmuggler to prevent premature templating from Jinja's internal dict copy."""
        return dict(_smuggled_vars=ArgSmuggler(jinja_vars=jinja_vars))

    @classmethod
    def extract_jinja_vars(cls, maybe_smuggled_vars: c.Mapping[str, t.Any] | None) -> c.Mapping[str, t.Any]:
        """
        If the supplied vars dict contains an ArgSmuggler instance with the expected key, unwrap it and return the smuggled value.
        Otherwise, return the supplied dict as-is.
        """
        if maybe_smuggled_vars and ((smuggler := maybe_smuggled_vars.get('_smuggled_vars')) and isinstance(smuggler, ArgSmuggler)):
            return smuggler.jinja_vars

        return maybe_smuggled_vars


class AnsibleTemplateExpression:
    """
    Wrapper around Jinja's TemplateExpression for converting MarkerError back into Marker.
    This is needed to make expression error handling consistent with templates, since Jinja does not support a custom type for Environment.compile_expression.
    """

    def __init__(self, template_expression: TemplateExpression) -> None:
        self._template_expression = template_expression

    def __call__(self, jinja_vars: c.Mapping[str, t.Any]) -> t.Any:
        try:
            return self._template_expression(ArgSmuggler.package_jinja_vars(jinja_vars))
        except MarkerError as ex:
            return ex.source


class AnsibleTemplate(Template):
    """
    A helper class, which prevents Jinja2 from running lazy containers through dict().
    """

    _python_source_temp_path: pathlib.Path | None = None

    def __del__(self):
        # DTFIX-FUTURE: this still isn't working reliably; something else must be keeping the template object alive
        if self._python_source_temp_path:
            self._python_source_temp_path.unlink(missing_ok=True)

    def __call__(self, jinja_vars: c.Mapping[str, t.Any]) -> t.Any:
        return self.render(ArgSmuggler.package_jinja_vars(jinja_vars))

    # noinspection PyShadowingBuiltins
    def new_context(
        self,
        vars: c.Mapping[str, t.Any] | None = None,
        shared: bool = False,
        locals: c.Mapping[str, t.Any] | None = None,
    ) -> Context:
        return _new_context(
            environment=self.environment,
            template_name=self.name,
            blocks=self.blocks,
            shared=shared,
            jinja_locals=locals,
            jinja_vars=ArgSmuggler.extract_jinja_vars(vars),
            jinja_globals=self.globals,
        )


class AnsibleCodeGenerator(NativeCodeGenerator):
    """
    Custom code generation behavior to support deprecated Ansible features and fill in gaps in Jinja native.
    This can be removed once the deprecated Ansible features are removed and the native fixes are upstreamed in Jinja.
    """

    def _output_const_repr(self, group: t.Iterable[t.Any]) -> str:
        """
        Prevent Jinja's code generation from stringifying single nodes before generating its repr.
        This complements the behavioral change in AnsibleEnvironment.concat which returns single nodes without stringifying them.
        """
        # DTFIX-FUTURE: contribute this upstream as a fix to Jinja's native support
        group_list = list(group)

        if len(group_list) == 1:
            return repr(group_list[0])

        # NB: This is slightly more efficient than Jinja's _output_const_repr, which generates a throw-away list instance to pass to join.
        #     Before removing this, ensure that upstream Jinja has this change.
        return repr("".join(map(str, group_list)))

    def visit_Const(self, node: Const, frame: Frame) -> None:
        """
        Override Jinja's visit_Const to inject a runtime call to AnsibleEnvironment._access_const for constant strings that are possibly templates, which
        may require special handling at runtime. See that method for details. An example that hits this path:
        {{ lookup("file", "{{ output_dir }}/bla") }}
        """
        value = node.as_const(frame.eval_ctx)

        if _TemplateConfig.allow_embedded_templates and type(value) is str and is_possibly_template(value):  # pylint: disable=unidiomatic-typecheck
            # deprecated: description='embedded Jinja constant string template support' core_version='2.23'
            self.write(f'environment._access_const({value!r})')
        else:
            # NB: This is actually more efficient than Jinja's visit_Const, which contains obsolete (as of Py2.7/3.1) float conversion instance checks. Before
            #     removing this override entirely, ensure that upstream Jinja has removed the obsolete code.
            #     See https://docs.python.org/release/2.7/whatsnew/2.7.html#python-3-1-features for more details.
            self.write(repr(value))


@pass_context
def _ansible_finalize(_ctx: AnsibleContext, value: t.Any) -> t.Any:
    """
    This function is called by Jinja with the result of each variable template block (e.g., {{ }}) encountered in a template.
    The pass_context decorator prevents finalize from being called on constants at template compile time.
    The passed in AnsibleContext is unused -- it is the result of using the pass_context decorator.
    The important part for us is that this blocks constant folding, which ensures our custom visit_Const is used.
    It also ensures that template results are wrapped in lazy containers.
    """
    return lazify_container(value)


@dataclasses.dataclass(kw_only=True, slots=True)
class _TemplateCompileContext(_ambient_context.AmbientContextBase):
    """
    This context is active during Ansible's explicit compilation of templates/expressions, but not during Jinja's runtime compilation.
    Historically, Ansible-specific pre-processing like `escape_backslashes` was not applied to imported/included templates.
    """

    escape_backslashes: bool


class _CompileStateSmugglingCtx(_ambient_context.AmbientContextBase):
    template_source: str | None = None
    python_source: str | None = None
    python_source_temp_path: pathlib.Path | None = None


class AnsibleLexer(Lexer):
    """
    Lexer override to escape backslashes in string constants within Jinja expressions; prevents Jinja from double-escaping them.

    NOTE: This behavior is only applied to string constants within Jinja expressions (eg {{ "c:\newfile" }}), *not* statements ("{% set foo="c:\\newfile" %}").

    This is useful when templates are sourced from YAML double-quoted strings, as it avoids having backslashes processed twice: first by the
    YAML parser, and then again by the Jinja parser. Instead, backslashes are only processed by YAML.

    Example YAML:

    - debug:
        msg: "Test Case 1\\3; {{ test1_name | regex_replace('^(.*)_name$', '\\1')}}"

    Since the outermost YAML string is double-quoted, the YAML parser converts the double backslashes to single backslashes. Without escaping, Jinja
    would see only a single backslash ('\1') while processing the embedded template expression, interpret it as an escape sequence, and convert it
    to '\x01' (ASCII "SOH"). This is clearly not the intended `\1` backreference argument to the `regex_replace` filter (which would require the
    double-escaped string '\\\\1' to yield the intended result).

    Since the "\\3" in the input YAML was not part of a template expression, the YAML-parsed "\3" remains after Jinja rendering. This would be
    confusing for playbook authors, as different escaping rules would be needed inside and outside the template expression.

    When templates are not sourced from YAML, escaping backslashes will prevent use of backslash escape sequences such as "\n" and "\t".

    See relevant Jinja lexer impl at e.g.: https://github.com/pallets/jinja/blob/3.1.2/src/jinja2/lexer.py#L646-L653.
    """

    def tokeniter(self, *args, **kwargs) -> t.Iterator[t.Tuple[int, str, str]]:
        """Pre-escape backslashes in expression ({{ }}) raw string constants before Jinja's Lexer.wrap() can interpret them as ASCII escape sequences."""
        token_stream = super().tokeniter(*args, **kwargs)

        # if we have no context, Jinja's doing a nested compile at runtime (eg, import/include); historically, no backslash escaping is performed
        if not (tcc := _TemplateCompileContext.current(optional=True)) or not tcc.escape_backslashes:
            yield from token_stream
            return

        in_variable = False

        for token in token_stream:
            token_type = token[1]

            if token_type == TOKEN_VARIABLE_BEGIN:
                in_variable = True
            elif token_type == TOKEN_VARIABLE_END:
                in_variable = False
            elif in_variable and token_type == TOKEN_STRING:
                token = token[0], token_type, token[2].replace('\\', '\\\\')

            yield token


def defer_template_error(ex: Exception, variable: t.Any, *, is_expression: bool) -> Marker:
    if not ex.__traceback__:
        raise AssertionError('ex must be a previously raised exception')

    if isinstance(ex, MarkerError):
        return ex.source

    exception_to_raise = create_template_error(ex, variable, is_expression)

    if exception_to_raise is ex:
        return CapturedExceptionMarker(ex)  # capture the previously raised exception

    try:
        raise exception_to_raise from ex  # raise the newly synthesized exception before capturing it
    except Exception as captured_ex:
        return CapturedExceptionMarker(captured_ex)


def create_template_error(ex: Exception, variable: t.Any, is_expression: bool) -> AnsibleTemplateError:
    if isinstance(ex, AnsibleTemplateError):
        exception_to_raise = ex
    else:
        kind = "expression" if is_expression else "template"
        ex_type = AnsibleTemplateError  # always raise an AnsibleTemplateError/subclass

        if isinstance(ex, RecursionError):
            msg = f"Recursive loop detected in {kind}."
        elif isinstance(ex, TemplateSyntaxError):
            msg = f"Syntax error in {kind}."

            if is_expression and is_possibly_template(variable):
                msg += " Template delimiters are not supported in expressions."

            ex_type = AnsibleTemplateSyntaxError
        else:
            msg = f"Error rendering {kind}."

        exception_to_raise = ex_type(msg, obj=variable)

    if exception_to_raise.obj is None:
        exception_to_raise.obj = TemplateContext.current().template_value

    # DTFIX-FUTURE: Look through the TemplateContext hierarchy to find the most recent non-template
    #   caller and use that for origin when no origin is available on obj. This could be useful for situations where the template
    #   was embedded in a plugin, or a plugin is otherwise responsible for losing the origin and/or trust. We can't just use the first
    #   non-template caller as that will lead to false positives for re-entrant calls (e.g. template plugins that call into templar).

    return exception_to_raise


_BUILTIN_FILTER_ALIASES: dict[str, str] = {}
_BUILTIN_TEST_ALIASES: dict[str, str] = {
    '!=': 'ne',
    '<': 'lt',
    '<=': 'le',
    '==': 'eq',
    '>': 'gt',
    '>=': 'ge',
}

_BUILTIN_FILTERS = filter_loader._wrap_funcs(defaults.DEFAULT_FILTERS, _BUILTIN_FILTER_ALIASES)
_BUILTIN_TESTS = test_loader._wrap_funcs(t.cast(dict[str, t.Callable], defaults.DEFAULT_TESTS), _BUILTIN_TEST_ALIASES)


class AnsibleEnvironment(SandboxedEnvironment):
    """
    Our custom environment, which simply allows us to override the class-level
    values for the Template and Context classes used by jinja2 internally.
    """

    context_class = AnsibleContext
    template_class = AnsibleTemplate
    code_generator_class = AnsibleCodeGenerator
    intercepted_binops = frozenset(('eq',))

    _allowed_unsafe_attributes: dict[str, type | tuple[type, ...]] = dict(
        # Allow bitwise operations on int until bitwise filters are available.
        # see: https://github.com/ansible/ansible/issues/85204
        __and__=int,
        __lshift__=int,
        __or__=int,
        __rshift__=int,
        __xor__=int,
    )
    """
    Attributes which are considered unsafe by `is_safe_attribute`, which should be allowed when used on specific types.
    The attributes allowed here are intended only for backward compatibility with existing use cases.
    They should be exposed as filters in a future release and eventually deprecated.
    """

    _lexer_cache = LRUCache(50)

    # DTFIX-FUTURE: bikeshed a name/mechanism to control template debugging
    _debuggable_template_source = False
    _debuggable_template_source_path: pathlib.Path = pathlib.Path(__file__).parent.parent.parent.parent / '.template_debug_source'

    def __init__(self, *args, ansible_basedir: str | None = None, **kwargs) -> None:
        if ansible_basedir:
            kwargs.update(loader=FileSystemLoader(ansible_basedir))

        super().__init__(*args, extensions=_TemplateConfig.jinja_extensions, **kwargs)

        self.filters = JinjaPluginIntercept(_BUILTIN_FILTERS, filter_loader)  # type: ignore[assignment]
        self.tests = JinjaPluginIntercept(_BUILTIN_TESTS, test_loader)  # type: ignore[assignment,arg-type]

        # future Jinja releases may default-enable autoescape; force-disable to prevent the problems it could cause
        # see https://github.com/pallets/jinja/blob/3.1.2/docs/api.rst?plain=1#L69
        self.autoescape = False

        self.trim_blocks = True

        self.undefined = UndefinedMarker
        self.finalize = _ansible_finalize

        self.globals.update(
            range=range,  # the sandboxed environment limits range in ways that may cause us problems; use the real Python one
            now=_now,
            undef=_undef,
            omit=Omit,
            lookup=_lookup,
            query=_query,
            q=_query,
        )

        # Disabling the optimizer prevents compile-time constant expression folding, which prevents our
        # visit_Const recursive inline template expansion tricks from working in many cases where Jinja's
        # ignorance of our embedded templates are optimized away as fully-constant expressions,
        # eg {{ "{{'hi'}}" == "hi" }}. As of Jinja ~3.1, this specifically avoids cases where the @optimizeconst
        # visitor decorator performs constant folding, which bypasses our visit_Const impl and causes embedded
        # templates to be lost.
        # See also optimizeconst impl: https://github.com/pallets/jinja/blob/3.1.0/src/jinja2/compiler.py#L48-L49
        self.optimized = False

    def get_template(
        self,
        name: str | Template,
        parent: str | None = None,
        globals: c.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        """Ensures that templates built via `get_template` are also source debuggable."""
        with _CompileStateSmugglingCtx.when(self._debuggable_template_source) as ctx:
            template_obj = t.cast(AnsibleTemplate, super().get_template(name, parent, globals))

            if isinstance(ctx, _CompileStateSmugglingCtx):  # only present if debugging is enabled
                template_obj._python_source_temp_path = ctx.python_source_temp_path  # facilitate deletion of the temp file when template_obj is deleted

            return template_obj

    def is_safe_attribute(self, obj: t.Any, attr: str, value: t.Any) -> bool:
        # deprecated: description="remove relaxed template sandbox mode support" core_version="2.23"
        if _TemplateConfig.sandbox_mode == _SandboxMode.ALLOW_UNSAFE_ATTRIBUTES:
            return True

        if (type_or_tuple := self._allowed_unsafe_attributes.get(attr)) and isinstance(obj, type_or_tuple):
            return True

        return super().is_safe_attribute(obj, attr, value)

    @property
    def lexer(self) -> AnsibleLexer:
        """Return/cache an AnsibleLexer with settings from the current AnsibleEnvironment"""
        # DTFIX-FUTURE: optimization - we should pre-generate the default cached lexer before forking, not leave it to chance (e.g. simple playbooks)
        key = tuple(getattr(self, name) for name in _TEMPLATE_OVERRIDE_FIELD_NAMES)

        lex = self._lexer_cache.get(key)

        if lex is None:
            self._lexer_cache[key] = lex = AnsibleLexer(self)

        return lex

    def call_filter(
        self,
        name: str,
        value: t.Any,
        args: c.Sequence[t.Any] | None = None,
        kwargs: c.Mapping[str, t.Any] | None = None,
        context: Context | None = None,
        eval_ctx: EvalContext | None = None,
    ) -> t.Any:
        """
        Ensure that filters directly invoked by plugins will see non-templating lazy containers.
        Without this, `_wrap_filter` will wrap `args` and `kwargs` in templating lazy containers.
        This provides consistency with plugin output handling by preventing auto-templating of trusted templates passed in native containers.
        """
        # DTFIX-FUTURE: need better logic to handle non-list/non-dict inputs for args/kwargs
        args = _AnsibleLazyTemplateMixin._try_create(list(args or []), LazyOptions.SKIP_TEMPLATES)
        kwargs = _AnsibleLazyTemplateMixin._try_create(kwargs, LazyOptions.SKIP_TEMPLATES)

        return super().call_filter(name, value, args, kwargs, context, eval_ctx)

    def call_test(
        self,
        name: str,
        value: t.Any,
        args: c.Sequence[t.Any] | None = None,
        kwargs: c.Mapping[str, t.Any] | None = None,
        context: Context | None = None,
        eval_ctx: EvalContext | None = None,
    ) -> t.Any:
        """
        Ensure that tests directly invoked by plugins will see non-templating lazy containers.
        Without this, `_wrap_test` will wrap `args` and `kwargs` in templating lazy containers.
        This provides consistency with plugin output handling by preventing auto-templating of trusted templates passed in native containers.
        """
        # DTFIX-FUTURE: need better logic to handle non-list/non-dict inputs for args/kwargs
        args = _AnsibleLazyTemplateMixin._try_create(list(args or []), LazyOptions.SKIP_TEMPLATES)
        kwargs = _AnsibleLazyTemplateMixin._try_create(kwargs, LazyOptions.SKIP_TEMPLATES)

        return super().call_test(name, value, args, kwargs, context, eval_ctx)

    def compile_expression(self, source: str, *args, **kwargs) -> TemplateExpression:
        # compile_expression parses and passes the tree to from_string; for debug support, activate the context here to capture the intermediate results
        with _CompileStateSmugglingCtx.when(self._debuggable_template_source) as ctx:
            if isinstance(ctx, _CompileStateSmugglingCtx):  # only present if debugging is enabled
                ctx.template_source = source

            return super().compile_expression(source, *args, **kwargs)

    def from_string(self, source: str | jinja2.nodes.Template, *args, **kwargs) -> AnsibleTemplate:
        # if debugging is enabled, use existing context when present (e.g., from compile_expression)
        current_ctx = _CompileStateSmugglingCtx.current(optional=True) if self._debuggable_template_source else None

        with _CompileStateSmugglingCtx.when(self._debuggable_template_source and not current_ctx) as new_ctx:
            template_obj = t.cast(AnsibleTemplate, super().from_string(source, *args, **kwargs))

            if isinstance(ctx := current_ctx or new_ctx, _CompileStateSmugglingCtx):  # only present if debugging is enabled
                template_obj._python_source_temp_path = ctx.python_source_temp_path  # facilitate deletion of the temp file when template_obj is deleted

        return template_obj

    def _parse(self, source: str, *args, **kwargs) -> jinja2.nodes.Template:
        if csc := _CompileStateSmugglingCtx.current(optional=True):
            csc.template_source = source

        return super()._parse(source, *args, **kwargs)

    def _compile(self, source: str, filename: str) -> types.CodeType:
        if csc := _CompileStateSmugglingCtx.current(optional=True):
            origin = Origin.get_tag(csc.template_source) or Origin.UNKNOWN

            source = '\n'.join(
                (
                    "import sys; breakpoint() if type(sys.breakpointhook) is not type(breakpoint) else None",
                    f"# original template source from {str(origin)!r}: ",
                    '\n'.join(f'# {line}' for line in (csc.template_source or '').splitlines()),
                    source,
                )
            )

            source_temp_dir = self._debuggable_template_source_path
            source_temp_dir.mkdir(parents=True, exist_ok=True)

            with tempfile.NamedTemporaryFile(dir=source_temp_dir, mode='w', suffix='.py', prefix='j2_src_', delete=False) as source_file:
                filename = source_file.name

                source_file.write(source)
                source_file.flush()

            csc.python_source = source
            csc.python_source_temp_path = pathlib.Path(filename)

        res = super()._compile(source, filename)

        return res

    @staticmethod
    def concat(nodes: t.Iterable[t.Any]) -> t.Any:  # type: ignore[override]
        node_list = list(_flatten_nodes(nodes))

        if not node_list:
            return None

        # this code is complemented by our tweaked CodeGenerator _output_const_repr that ensures that literal constants
        # in templates aren't double-repr'd in the generated code
        if len(node_list) == 1:
            return node_list[0]

        # In order to ensure that all markers are tripped, do a recursive finalize before we repr (otherwise we can end up
        # repr'ing a Marker). This requires two passes, but avoids the need for a parallel reimplementation of all repr methods.
        try:
            node_list = _finalize_template_result(node_list, FinalizeMode.CONCAT)
        except MarkerError as ex:
            return ex.source  # return the first Marker encountered

        return ''.join([to_text(v) for v in node_list])

    @staticmethod
    def _access_const(const_template: t.LiteralString) -> t.Any:
        """
        Called during template rendering on template-looking string constants embedded in the template.
        It provides the following functionality:
        * Propagates origin from the containing template.
        * For backward compatibility when embedded templates are enabled:
          * Conditionals - Renders embedded template constants and accesses the result. Warns on each constant immediately.
          * Non-conditionals - Tags constants for deferred rendering of templates in lookup terms. Warns on each constant during lookup invocation.
        """
        ctx = TemplateContext.current()

        if (tv := ctx.template_value) and (origin := Origin.get_tag(tv)):
            const_template = origin.tag(const_template)

        if ctx._render_jinja_const_template:
            _jinja_const_template_warning(const_template, is_conditional=True)

            result = ctx.templar.template(TrustedAsTemplate().tag(const_template))
            AnsibleAccessContext.current().access(result)
        else:
            # warnings will be issued when lookup terms processing occurs, to avoid false positives
            result = _JinjaConstTemplate().tag(const_template)

        return result

    def getitem(self, obj: t.Any, argument: t.Any) -> t.Any:
        value = super().getitem(obj, argument)

        AnsibleAccessContext.current().access(value)

        return value

    def getattr(self, obj: t.Any, attribute: str) -> t.Any:
        """
        Get `attribute` from the attributes of `obj`, falling back to items in `obj`.
        If no item was found, return a sandbox-specific `UndefinedMarker` if `attribute` is protected by the sandbox,
        otherwise return a normal `UndefinedMarker` instance.
        This differs from the built-in Jinja behavior which will not fall back to items if `attribute` is protected by the sandbox.
        """
        # example template that uses this: "{{ some.thing }}" -- obj is the "some" dict, attribute is "thing"

        is_safe = True

        try:
            value = getattr(obj, attribute)
        except AttributeError:
            value = _sentinel
        else:
            if not (is_safe := self.is_safe_attribute(obj, attribute, value)):
                value = _sentinel

        if value is _sentinel:
            try:
                value = obj[attribute]
            except (TypeError, LookupError):
                value = self.undefined(obj=obj, name=attribute) if is_safe else self.unsafe_undefined(obj, attribute)

        AnsibleAccessContext.current().access(value)

        return value

    def call(
        self,
        __context: Context,
        __obj: t.Any,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> t.Any:
        try:
            if _DirectCall.is_marked(__obj):
                # Both `_lookup` and `_query` handle arg proxying and `Marker` args internally.
                # Performing either before calling them will interfere with that processing.
                return super().call(__context, __obj, *args, **kwargs)

            # Jinja's generated macro code handles Markers, so preemptive raise on Marker args and lazy retrieval should be disabled for the macro invocation.
            is_macro = isinstance(__obj, Macro)

            if not is_macro and (first_marker := get_first_marker_arg(args, kwargs)) is not None:
                return first_marker

            with JinjaCallContext(accept_lazy_markers=is_macro):
                call_res = super().call(__context, __obj, *lazify_container_args(args), **lazify_container_kwargs(kwargs))

                if __obj is range:
                    # Preserve the ability to do `range(1000000000) | random` by not converting range objects to lists.
                    # Historically, range objects were only converted on Jinja finalize and filter outputs, so they've always been floating around in templating
                    # code and visible to user plugins.
                    return call_res

                return _wrap_plugin_output(call_res)

        except MarkerError as ex:
            return ex.source
        except Exception as ex:
            return CapturedExceptionMarker(ex)


AnsibleTemplate.environment_class = AnsibleEnvironment

_DEFAULT_UNDEF = UndefinedMarker("Mandatory variable has not been overridden", _no_template_source=True)

_sentinel: t.Final[object] = object()


@_DirectCall.mark
def _undef(hint: str | None = None) -> UndefinedMarker:
    """Jinja2 global function (undef) for creating getting a `UndefinedMarker` instance, optionally with a custom hint."""
    validate_arg_type('hint', hint, (str, type(None)))

    if not hint:
        return _DEFAULT_UNDEF

    return UndefinedMarker(hint)


def _flatten_nodes(nodes: t.Iterable[t.Any]) -> t.Iterable[t.Any]:
    """
    Yield nodes from a potentially recursive iterable of nodes.
    The recursion is required to expand template imports (TemplateModule).
    Any exception raised while consuming a template node will be yielded as a Marker for that node.
    """
    iterator = iter(nodes)

    while True:
        try:
            node = next(iterator)
        except StopIteration:
            break
        except Exception as ex:
            yield defer_template_error(ex, TemplateContext.current().template_value, is_expression=False)
            # DTFIX-FUTURE: We should be able to determine if truncation occurred by having the code generator smuggle out the number of expected nodes.
            yield TruncationMarker()
        else:
            if type(node) is TemplateModule:  # pylint: disable=unidiomatic-typecheck
                yield from _flatten_nodes(node._body_stream)
            elif node is None:
                continue  # avoid yielding `None`-valued nodes to avoid literal "None" in stringified template results
            else:
                yield node


def _flatten_and_lazify_vars(mapping: c.Mapping) -> t.Iterable[c.Mapping]:
    """Prevent deeply-nested Jinja vars ChainMaps from being created by nested contexts and ensure that all top-level containers support lazy templating."""
    mapping_type = type(mapping)
    if mapping_type is ChainMap:
        # noinspection PyUnresolvedReferences
        for m in mapping.maps:
            yield from _flatten_and_lazify_vars(m)
    elif mapping_type is _AnsibleLazyTemplateDict:
        yield mapping
    elif mapping_type in (dict, _AnsibleTaggedDict):
        # don't propagate empty dictionary layers
        if mapping:
            yield _AnsibleLazyTemplateMixin._try_create(mapping)
    else:
        raise NotImplementedError(f"unsupported mapping type in Jinja vars: {mapping_type}")


def _new_context(
    *,
    environment: Environment,
    template_name: str | None,
    blocks: dict[str, t.Callable[[Context], c.Iterator[str]]],
    shared: bool = False,
    jinja_locals: c.Mapping[str, t.Any] | None = None,
    jinja_vars: c.Mapping[str, t.Any] | None = None,
    jinja_globals: c.MutableMapping[str, t.Any] | None = None,
) -> Context:
    """Override Jinja's context vars setup to use ChainMaps and containers that support lazy templating."""
    layers = []

    if jinja_locals:
        # Omit values set to Jinja's internal `missing` sentinel; they are locals that have not yet been
        # initialized in the current context, and should not be exposed to child contexts. e.g.: {% import 'a' as b with context %}.
        # The `b` local will be `missing` in the `a` context and should not be propagated as a local to the child context we're creating.
        layers.append(_AnsibleLazyTemplateMixin._try_create({k: v for k, v in jinja_locals.items() if v is not missing}))

    if jinja_vars:
        layers.extend(_flatten_and_lazify_vars(jinja_vars))

    if jinja_globals and not shared:
        # Even though we don't currently support templating globals, it's easier to ensure that everything is template-able rather than trying to
        # pick apart the ChainMaps to enforce non-template-able globals, or to risk things that *should* be template-able not being lazified.
        layers.extend(_flatten_and_lazify_vars(jinja_globals))

    if not layers:
        # ensure we have at least one layer (which should be lazy), since _flatten_and_lazify_vars eliminates most empty layers
        layers.append(_AnsibleLazyTemplateMixin._try_create({}))

    # only return a ChainMap if we're combining layers, or we have none
    parent = layers[0] if len(layers) == 1 else ChainMap(*layers)

    # the `parent` cast is only to satisfy Jinja's overly-strict type hint
    return environment.context_class(environment, t.cast(dict, parent), template_name, blocks, globals=jinja_globals)


def is_possibly_template(value: str, overrides: TemplateOverrides = TemplateOverrides.DEFAULT):
    """
    A lightweight check to determine if the given string looks like it contains a template, even if that template is invalid.
    Returns `True` if the given string starts with a Jinja overrides header or if it contains template start strings.
    """
    return value.startswith(JINJA2_OVERRIDE) or overrides._contains_start_string(value)


def is_possibly_all_template(value: str, overrides: TemplateOverrides = TemplateOverrides.DEFAULT):
    """
    A lightweight check to determine if the given string looks like it contains *only* a template, even if that template is invalid.
    Returns `True` if the given string starts with a Jinja overrides header or if it starts and ends with Jinja template delimiters.
    """
    return value.startswith(JINJA2_OVERRIDE) or overrides._starts_and_ends_with_jinja_delimiters(value)


class FinalizeMode(enum.Enum):
    TOP_LEVEL = enum.auto()
    CONCAT = enum.auto()


_FINALIZE_FAST_PATH_EXACT_MAPPING_TYPES = frozenset(
    (
        dict,
        _AnsibleTaggedDict,
        _AnsibleLazyTemplateDict,
        HostVars,
        HostVarsVars,
    )
)
"""Fast-path exact mapping types for finalization. These types bypass diagnostic warnings for type conversion."""

_FINALIZE_FAST_PATH_EXACT_ITERABLE_TYPES = frozenset(
    (
        list,
        _AnsibleTaggedList,
        _AnsibleLazyTemplateList,
        tuple,
        _AnsibleTaggedTuple,
        _AnsibleLazyAccessTuple,
    )
)
"""Fast-path exact iterable types for finalization. These types bypass diagnostic warnings for type conversion."""

_FINALIZE_DISALLOWED_EXACT_TYPES = frozenset((range,))
"""Exact types that cannot be finalized."""

# Jinja passes these into filters/tests via @pass_environment
register_known_types(
    AnsibleContext,
    AnsibleEnvironment,
    EvalContext,
)


def _finalize_dict(o: t.Any, mode: FinalizeMode) -> t.Iterator[tuple[t.Any, t.Any]]:
    for k, v in o.items():
        if v is not Omit:
            yield _finalize_template_result(k, mode), _finalize_template_result(v, mode)


def _finalize_list(o: t.Any, mode: FinalizeMode) -> t.Iterator[t.Any]:
    for v in o:
        if v is not Omit:
            yield _finalize_template_result(v, mode)


def _maybe_finalize_scalar(o: t.Any) -> t.Any:
    # DTFIX5: this should check all supported scalar subclasses, not just JSON ones (also, does the JSON serializer handle these cases?)
    for target_type in _json_subclassable_scalar_types:
        if not isinstance(o, target_type):
            continue

        match _TemplateConfig.unknown_type_conversion_handler.action:
            # we don't want to show the object value, and it can't be Origin-tagged; send the current template value for best effort
            case ErrorAction.WARNING:
                display.warning(
                    msg=f'Type {native_type_name(o)!r} is unsupported in variable storage, converting to {native_type_name(target_type)!r}.',
                    obj=TemplateContext.current(optional=True).template_value,
                )
            case ErrorAction.ERROR:
                raise AnsibleVariableTypeError.from_value(obj=TemplateContext.current(optional=True).template_value)

        return target_type(o)

    return None


def _finalize_fallback_collection(
    o: t.Any,
    mode: FinalizeMode,
    finalizer: t.Callable[[t.Any, FinalizeMode], t.Iterator],
    target_type: type[list | dict],
) -> t.Collection[t.Any]:
    match _TemplateConfig.unknown_type_conversion_handler.action:
        # we don't want to show the object value, and it can't be Origin-tagged; send the current template value for best effort
        case ErrorAction.WARNING:
            display.warning(
                msg=f'Type {native_type_name(o)!r} is unsupported in variable storage, converting to {native_type_name(target_type)!r}.',
                obj=TemplateContext.current(optional=True).template_value,
            )
        case ErrorAction.ERROR:
            raise AnsibleVariableTypeError.from_value(obj=TemplateContext.current(optional=True).template_value)

    return _finalize_collection(o, mode, finalizer, target_type)


def _finalize_collection(
    o: t.Any,
    mode: FinalizeMode,
    finalizer: t.Callable[[t.Any, FinalizeMode], t.Iterator],
    target_type: type[list | dict],
) -> t.Collection[t.Any]:
    return AnsibleTagHelper.tag(finalizer(o, mode), AnsibleTagHelper.tags(o), value_type=target_type)


def _finalize_template_result(o: t.Any, mode: FinalizeMode) -> t.Any:
    """Recurse the template result, rendering any encountered templates, converting containers to non-lazy versions."""
    # DTFIX5: add tests to ensure this method doesn't drift from allowed types
    o_type = type(o)

    # DTFIX-FUTURE: provide an optional way to check for trusted templates leaking out of templating (injected, but not passed through templar.template)

    if o_type is _AnsibleTaggedStr:
        return _JinjaConstTemplate.untag(o)  # prevent _JinjaConstTemplate from leaking into finalized results

    if o_type in PASS_THROUGH_SCALAR_VAR_TYPES:
        return o

    if o_type in _FINALIZE_FAST_PATH_EXACT_MAPPING_TYPES:  # silently convert known mapping types to dict
        return _finalize_collection(o, mode, _finalize_dict, dict)

    if o_type in _FINALIZE_FAST_PATH_EXACT_ITERABLE_TYPES:  # silently convert known sequence types to list
        return _finalize_collection(o, mode, _finalize_list, list)

    if o_type in Marker._concrete_subclasses:  # this early return assumes handle_marker follows our variable type rules
        return TemplateContext.current().templar.marker_behavior.handle_marker(o)

    if mode is not FinalizeMode.TOP_LEVEL:  # unsupported type (do not raise)
        return o

    if o_type in _FINALIZE_DISALLOWED_EXACT_TYPES:  # early abort for disallowed types that would otherwise be handled below
        raise AnsibleVariableTypeError.from_value(obj=o)

    if _internal.is_intermediate_mapping(o):  # since isinstance checks are slower, this is separate from the exact type check above
        return _finalize_fallback_collection(o, mode, _finalize_dict, dict)

    if _internal.is_intermediate_iterable(o):  # since isinstance checks are slower, this is separate from the exact type check above
        return _finalize_fallback_collection(o, mode, _finalize_list, list)

    if (result := _maybe_finalize_scalar(o)) is not None:
        return result

    raise AnsibleVariableTypeError.from_value(obj=o)



# DTFIX5: review this test for removal, should have been superseded by test_serialization_profiles.py
#  some tests will need to be under module_utils to verify target-only Python versions
#  if kept, de-dupe
from __future__ import annotations

import json
import typing as t
import collections.abc as c

import pytest

from ansible.module_utils._internal._datatag import AnsibleTaggedObject, _untaggable_types, AnsibleTagHelper
from ansible.module_utils._internal._datatag._tags import Deprecated
from ansible._internal._datatag._tags import Origin, TrustedAsTemplate
from ansible.module_utils.common.json import get_encoder, get_decoder
from ansible.module_utils._internal._json._profiles import _module_legacy_m2c, _tagless, _module_legacy_c2m, _module_modern_c2m, _module_modern_m2c
from ansible._internal._json._profiles import _legacy

_simple_json_values = (
    'hello',
    1.1,
    1,
    True,
    None,
    [1],
    dict(k='v'),
)

_converted_json_values = (
    ((1,), [1]),
    ({1}, [1]),
)

origin = Origin(path='/absolute/path/for/testing')


@pytest.mark.parametrize("value, expected", _converted_json_values, ids=str)
def test_modern_controller_to_module_converted_types(value: t.Any, expected: t.Any) -> None:
    args = dict(untagged_value=value, tagged_value=origin.tag(value))
    profile = _module_modern_c2m
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']

    assert untagged_value == expected
    assert tagged_value == expected

    assert_has_no_tags(result)


@pytest.mark.parametrize("value, expected", _converted_json_values, ids=str)
def test_legacy_controller_to_module_converted_types(value: t.Any, expected: t.Any) -> None:
    args = dict(untagged_value=value, tagged_value=origin.tag(value))
    profile = _module_legacy_c2m
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']

    assert untagged_value == expected
    assert tagged_value == expected

    assert_has_no_tags(result)


@pytest.mark.parametrize("value, expected", _converted_json_values, ids=str)
def test_modern_module_to_controller_converted_types(value: t.Any, expected: t.Any) -> None:
    deprecation_tag = Deprecated(msg='go away')
    args = dict(untagged_value=value, tagged_value=deprecation_tag.tag(value))
    profile = _module_modern_m2c
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']

    assert untagged_value == expected
    assert tagged_value == expected

    assert_has_no_tags(untagged_value)

    if type(value) not in _untaggable_types:
        tags = AnsibleTagHelper.tags(tagged_value)

        assert tags == {deprecation_tag}


@pytest.mark.parametrize("value, expected", _converted_json_values, ids=str)
def test_legacy_module_to_controller_converted_types(value: t.Any, expected: t.Any) -> None:
    args = dict(untagged_value=value)
    profile = _module_modern_m2c
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']

    assert untagged_value == expected

    assert_has_no_tags(result)


@pytest.mark.parametrize("value", _simple_json_values, ids=str)
def test_modern_controller_to_module(value: t.Any) -> None:
    args = dict(untagged_value=value, tagged_value=origin.tag(value))
    profile = _module_modern_c2m
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']

    assert untagged_value == value
    assert tagged_value == value

    assert_has_no_tags(result)


@pytest.mark.parametrize("value", _simple_json_values, ids=str)
def test_legacy_controller_to_module(value: t.Any) -> None:
    args = dict(untagged_value=value, tagged_value=origin.tag(value))
    profile = _module_legacy_c2m
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']

    assert untagged_value == value
    assert tagged_value == value

    assert_has_no_tags(result)


@pytest.mark.parametrize("value", _simple_json_values, ids=str)
def test_modern_module_to_controller(value: t.Any) -> None:
    deprecation_tag = Deprecated(msg='go away')
    args = dict(
        untagged_value=value,
        tagged_value=deprecation_tag.tag(value),
        # tagged_with_unwanted_tag_only=origin.tag(value),
    )
    profile = _module_modern_m2c
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']
    # tagged_with_unwanted_tag_only = result['tagged_with_unwanted_tag_only']

    assert untagged_value == value
    assert tagged_value == value
    # assert tagged_with_unwanted_tag_only == value

    assert_has_no_tags(untagged_value)
    # assert_has_no_tags(tagged_with_unwanted_tag_only)

    if type(value) not in _untaggable_types:
        tags = AnsibleTagHelper.tags(tagged_value)

        assert tags == {deprecation_tag}


@pytest.mark.parametrize("value", _simple_json_values, ids=str)
def test_legacy_module_to_controller(value: t.Any) -> None:
    # deprecation_tag = Deprecated(msg='go away')
    args = dict(
        untagged_value=value,
        # tagged_value=deprecation_tag.tag(value),
        # tagged_with_unwanted_tag_only=origin.tag(value),
    )
    profile = _module_legacy_m2c
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    # tagged_value = result['tagged_value']
    # tagged_with_unwanted_tag_only = result['tagged_with_unwanted_tag_only']

    assert untagged_value == value
    # assert tagged_value == value
    # assert tagged_with_unwanted_tag_only == value

    assert_has_no_tags(untagged_value)
    # assert_has_no_tags(tagged_value)
    # assert_has_no_tags(tagged_with_unwanted_tag_only)


@pytest.mark.parametrize("value", _simple_json_values, ids=str)
def test_legacy(value: t.Any) -> None:
    args = dict(
        untagged_value=value,
        tagged_value=origin.tag(value),
        trusted_value=apply_trust(value),
    )
    profile = _legacy
    payload = json.dumps(args, cls=get_encoder(profile))
    payload = TrustedAsTemplate().tag(payload)
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']
    trusted_value = result['trusted_value']

    assert untagged_value == value
    assert tagged_value == value
    assert trusted_value == value

    assert_has_no_tags(untagged_value)
    assert_has_no_tags(tagged_value)
    assert_has_trusted_strings(trusted_value)


@pytest.mark.parametrize("value, expected", (
    ((1,), [1]),
    ({1}, [1]),
), ids=str)
def test_legacy_converted_types(value: t.Any, expected: t.Any) -> None:
    args = dict(untagged_value=value, tagged_value=origin.tag(value))
    profile = _legacy
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']

    assert untagged_value == expected
    assert tagged_value == expected

    assert_has_no_tags(result)


@pytest.mark.parametrize("value", _simple_json_values, ids=str)
def test_tagless(value: t.Any) -> None:
    args = dict(
        untagged_value=value,
        tagged_value=origin.tag(value),
        trusted_value=apply_trust(value),
    )
    profile = _tagless
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']
    trusted_value = result['trusted_value']

    assert untagged_value == value
    assert tagged_value == value
    assert trusted_value == value

    assert_has_no_tags(result)


@pytest.mark.parametrize("value, expected", _converted_json_values, ids=str)
def test_tagless_converted_types(value: t.Any, expected: t.Any) -> None:
    args = dict(untagged_value=value, tagged_value=origin.tag(value))
    profile = _tagless
    payload = json.dumps(args, cls=get_encoder(profile))
    result = json.loads(payload, cls=get_decoder(profile))

    untagged_value = result['untagged_value']
    tagged_value = result['tagged_value']

    assert untagged_value == expected
    assert tagged_value == expected

    assert_has_no_tags(result)


def apply_trust(o: t.Any) -> t.Any:
    if isinstance(o, str):
        return TrustedAsTemplate().tag(o)

    if isinstance(o, dict):
        return {key: apply_trust(value) for key, value in o.items()}

    if isinstance(o, (set, list, tuple)):
        return type(o)(apply_trust(item) for item in o)

    return o


def assert_has_trusted_strings(o: t.Any) -> None:
    assert not isinstance(o, str) or TrustedAsTemplate.is_tagged_on(o)

    if isinstance(o, c.Mapping):
        for key, value in o.items():
            assert_has_no_tags(key)
            assert_has_trusted_strings(value)
    elif not isinstance(o, str) and isinstance(o, c.Iterable):
        for item in o:
            assert_has_trusted_strings(item)


def assert_has_no_tags(o: t.Any) -> None:
    assert not isinstance(o, AnsibleTaggedObject)

    if isinstance(o, c.Mapping):
        for key, value in o.items():
            assert_has_no_tags(key)
            assert_has_no_tags(value)
    elif not isinstance(o, str) and isinstance(o, c.Iterable):
        for item in o:
            assert_has_no_tags(item)



#!/usr/bin/python
from __future__ import annotations

import os

from ansible.module_utils.basic import AnsibleModule


def main():
    # This module verifies that AnsibleModule works when cwd does not exist.
    # This situation can occur as a race condition when the following conditions are met:
    #
    # 1) Execute a module which has high startup overhead prior to instantiating AnsibleModule (0.5s is enough in many cases).
    # 2) Run the module async as the last task in a playbook using connection=local (a fire-and-forget task).
    # 3) Remove the directory containing the playbook immediately after playbook execution ends (playbook in a temp dir).
    #
    # To ease testing of this race condition the deletion of cwd is handled in this module.
    # This avoids race conditions in the test, including timing cwd deletion between AnsiballZ wrapper execution and AnsibleModule instantiation.
    # The timing issue with AnsiballZ is due to cwd checking in the wrapper when code coverage is enabled.

    temp = os.path.abspath('temp')

    os.mkdir(temp)
    os.chdir(temp)
    os.rmdir(temp)

    module = AnsibleModule(argument_spec=dict())
    module.exit_json(before=temp, after=os.getcwd())


if __name__ == '__main__':
    main()



"""Legacy wire format for controller to module communication."""

from __future__ import annotations as _annotations

import datetime as _datetime

from .. import _profiles


class _Profile(_profiles._JSONSerializationProfile["Encoder", "Decoder"]):
    @classmethod
    def post_init(cls) -> None:
        cls.serialize_map = {}
        cls.serialize_map.update(cls._common_discard_tags)
        cls.serialize_map.update(
            {
                set: cls.serialize_as_list,  # legacy _json_encode_fallback behavior
                tuple: cls.serialize_as_list,  # JSONEncoder built-in behavior
                _datetime.date: cls.serialize_as_isoformat,
                _datetime.time: cls.serialize_as_isoformat,  # always failed pre-2.18, so okay to include for consistency
                _datetime.datetime: cls.serialize_as_isoformat,
            }
        )

        cls.handle_key = cls._handle_key_str_fallback  # type: ignore[method-assign]  # legacy stdlib-compatible key behavior


class Encoder(_profiles.AnsibleProfileJSONEncoder):
    _profile = _Profile


class Decoder(_profiles.AnsibleProfileJSONDecoder):
    _profile = _Profile



from __future__ import annotations

import os

from ansible.plugins.action import ActionBase
from ansible._internal._ssh._ssh_agent import SshAgentClient

from cryptography.hazmat.primitives.serialization import ssh


class ActionModule(ActionBase):

    def run(self, tmp=None, task_vars=None):
        results = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect
        match self._task.args['action']:
            case 'list':
                return self.list()
            case 'remove':
                return self.remove(self._task.args['pubkey'])
            case 'remove_all':
                return self.remove_all()
            case _:
                return {'failed': True, 'msg': 'not implemented'}

    def remove(self, pubkey_data):
        with SshAgentClient(os.environ['SSH_AUTH_SOCK']) as client:
            public_key = ssh.load_ssh_public_key(pubkey_data.encode())
            client.remove(public_key)
            return {'failed': public_key in client}

    def remove_all(self):
        with SshAgentClient(os.environ['SSH_AUTH_SOCK']) as client:
            nkeys_before = client.list().nkeys
            client.remove_all()
            nkeys_after = client.list().nkeys
            return {
                'failed': nkeys_after != 0,
                'nkeys_removed': nkeys_before,
            }

    def list(self):
        result = {'keys': [], 'nkeys': 0}
        with SshAgentClient(os.environ['SSH_AUTH_SOCK']) as client:
            key_list = client.list()
            result['nkeys'] = key_list.nkeys
            for key in key_list.keys:
                public_key = key.public_key
                key_size = getattr(public_key, 'key_size', 256)
                fingerprint = key.fingerprint
                key_type = key.type.main_type
                result['keys'].append({
                    'type': key_type,
                    'key_size': key_size,
                    'fingerprint': f'SHA256:{fingerprint}',
                    'comments': key.comments,
                })

        return result



"""Show availability of PyYAML and libyaml support."""

from __future__ import annotations

import json

try:
    import yaml
except ImportError:
    yaml = None

try:
    from yaml import CLoader
except ImportError:
    CLoader = None  # type: ignore[misc]

print(json.dumps(dict(
    yaml=bool(yaml),
    cloader=bool(CLoader),
)))



from __future__ import annotations

import pytest

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common import warnings

from ansible.module_utils.common.warnings import error_as_warning
from ansible.module_utils.testing import patch_module_args

pytestmark = pytest.mark.usefixtures("as_target", "module_env_mocker")


def test_error_as_warning() -> None:
    try:
        raise Exception('hello')
    except Exception as ex:
        error_as_warning('Warning message', ex)

    assert warnings.get_warning_messages() == ('Warning message: hello',)
    assert len(warnings.get_warnings()) == 1


def test_error_as_warning_via_module() -> None:
    with patch_module_args():
        am = AnsibleModule(argument_spec={})

    try:
        raise Exception('hello')
    except Exception as ex:
        am.error_as_warning('Warning message', ex)

    assert warnings.get_warning_messages() == ('Warning message: hello',)
    assert len(warnings.get_warnings()) == 1



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import os
import pathlib

from ansible.module_utils.facts.hardware import linux


def test_get_sysinfo_facts(monkeypatch):
    fixtures = pathlib.Path(__file__).parent / 'fixtures'
    sysinfo = (fixtures / 'sysinfo').read_text()

    monkeypatch.setattr(os.path, 'exists', lambda x: True)
    monkeypatch.setattr(linux, 'get_file_content', lambda x: sysinfo)

    lh = linux.LinuxHardware(None)
    facts = lh.get_sysinfo_facts()
    expected = {
        'system_vendor': 'IBM',
        'product_version': 'NA',
        'product_name': '8561',
        'product_serial': 'AB1CD',
        'product_uuid': 'NA',
    }
    assert facts == expected



from __future__ import annotations

import json
import re

import pytest

from ansible._internal._json._legacy_encoder import LegacyControllerJSONEncoder
from ansible.parsing.vault import EncryptedString, VaultLib, VaultSecretsContext


@pytest.fixture
def data(request: pytest.FixtureRequest, _vault_secrets_context) -> object:
    if not isinstance(request.param, EncryptedString):
        return request.param

    secret = VaultSecretsContext.current().secrets[0][1]
    ciphertext = VaultLib().encrypt(request.param._ciphertext.encode(), secret).decode()
    return EncryptedString(ciphertext=ciphertext)


@pytest.mark.parametrize("data, kw, expected", (
    ("astring", {}, '"astring"'),
    ("astring", dict(preprocess_unsafe=True), '{"__ansible_unsafe": "astring"}'),
    (EncryptedString(ciphertext="encrypted"), dict(vault_to_text=True), '"encrypted"'),
    (EncryptedString(ciphertext="encrypted"), {}, '{"__ansible_vault": "$ANSIBLE_VAULT...'),
    (b'wasbytes', dict(_decode_bytes=True), '"wasbytes"'),
), indirect=['data'])
def test_legacy_encoder_conversions(data: object, kw: dict[str, object], expected: str, _vault_secrets_context) -> None:
    if expected.endswith('...'):
        expected_pattern = re.escape(expected[:-3])
    else:
        expected_pattern = re.escape(expected) + "$"

    assert re.match(expected_pattern, json.dumps(data, cls=LegacyControllerJSONEncoder, **kw))  # type:ignore[arg-type]



from __future__ import annotations

import json
import pickle

from ansible._internal._datatag._tags import Origin
from ansible.utils.jsonrpc import JsonRpcServer


def test_response_type_cleansing() -> None:
    """Avoid unpickling errors in module contexts by ensuring that non-scalar JsonRpc responses are not pickled with tags."""

    class RPCTest:
        def returns_list_with_tagged_str(self) -> list:
            return [Origin(description="blar").tag("taggedstr")]

    s = JsonRpcServer()
    s.register(RPCTest())
    req = dict(method="returns_list_with_tagged_str", id=1, params=(tuple(), {}))
    jsonrpc_res = s.handle_request(json.dumps(req))

    deserialized_res = json.loads(jsonrpc_res)

    pickled_res = deserialized_res.get("result")

    assert pickled_res is not None

    res = pickle.loads(pickled_res.encode(errors="surrogateescape"))

    assert res == ["taggedstr"]
    assert not Origin.is_tagged_on(res[0])



# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import sys

import pytest

import ansible.module_utils.basic

from ansible.module_utils.testing import patch_module_args
from ..mock.module import module_env_mocker  # expose shared fixture in this part of the unit test tree

assert module_env_mocker is not None  # avoid unused imports


@pytest.fixture
def stdin(request):
    old_argv = sys.argv
    sys.argv = ['ansible_unittest']

    try:
        args = request.param.copy()
    except AttributeError:
        args = {}

    args.setdefault('_ansible_remote_tmp', '/tmp')
    args.setdefault('_ansible_keep_remote_files', False)
    args.setdefault('_ansible_tracebacks_for', [])

    with patch_module_args(args):
        yield

    sys.argv = old_argv


@pytest.fixture
def am(stdin, request):
    old_argv = sys.argv
    sys.argv = ['ansible_unittest']

    argspec = {}
    if hasattr(request, 'param'):
        if isinstance(request.param, dict):
            argspec = request.param

    am = ansible.module_utils.basic.AnsibleModule(
        argument_spec=argspec,
    )
    am._name = 'ansible_unittest'

    yield am

    sys.argv = old_argv
